"""
A collection of function definitions to handle common
thermodynamic calculations for the atmosphere.
"""

from __future__ import division
from numpy import exp, log
from .constants import (ps, kappa, tempCtoK, eps, Rd, Rv, cpv, cp, g, Lhvap,
                        sigma, hPlanck, c_light, kBoltzmann, molecular_weight)

def potential_temperature(T,p):
    """Compute potential temperature for an air parcel.

    Input:  T is temperature in Kelvin
            p is pressure in mb or hPa
    Output: potential temperature in Kelvin.

    """
    theta = T*(ps/p)**kappa
    return theta

def theta(T,p):
    '''Convenience method, identical to thermo.potential_temperature(T,p).'''
    return potential_temperature(T,p)

def temperature_from_potential(theta,p):
    """Convert potential temperature to in-situ temperature.

    Input:  theta is potential temperature in Kelvin
            p is pressure in mb or hPa
    Output: absolute temperature in Kelvin.

    """
    T = theta/((ps/p)**kappa)
    return T

def T(theta,p):
    '''Convenience method, identical to thermo.temperature_from_potential(theta,p).'''
    return temperature_from_potential(theta,p)

def clausius_clapeyron(T):
    """Compute saturation vapor pressure as function of temperature T.

    Input: T is temperature in Kelvin
    Output: saturation vapor pressure in mb or hPa

    Formula from Rogers and Yau "A Short Course in Cloud Physics" (Pergammon Press), p. 16
    claimed to be accurate to within 0.1% between -30degC and 35 degC
    Based on the paper by Bolton (1980, Monthly Weather Review).

    """
    Tcel = T - tempCtoK
    es = 6.112 * exp(17.67*Tcel/(Tcel+243.5))
    return es

def qsat(T,p):
    """Compute saturation specific humidity as function of temperature and pressure.

    Input:  T is temperature in Kelvin
            p is pressure in hPa or mb
    Output: saturation specific humidity (dimensionless).

    """
    es = clausius_clapeyron(T)
    q = eps * es / (p - (1 - eps) * es )
    return q

def virtual_temperature_from_mixing_ratio(T,w):
    '''Virtual temperature Tv
    T is air temperature (K)
    w is water vapor mixing ratio (dimensionless)'''
    return T * ((1+w/eps)/(1+w))

def vapor_pressure_from_specific_humidity(p,q):
    '''Vapor pressure (same units as input p)
    p is total air pressure
    q is specific humidity (dimensionless) -- mass of vapor per unit mass moist air'''
    return p * (q/(eps+q*(1-eps)))

def mixing_ratio_from_vapor_pressure(p,e):
    '''Water vapor mixing ratio
    p is air pressure
    e is vapor pressure
    p and e must be in same units (e.g. hPa)
    '''
    return eps * e / (p-e)

def rho_moist(T,p,q):
    '''Density of moist air.
    T is air temperature (K)
    p is air pressure (hPa)
    q is specific humidity (hPa)

    returns density in kg/m3
    '''
    e = vapor_pressure_from_specific_humidity(p,q)
    w = mixing_ratio_from_vapor_pressure(p,e)
    Tv = virtual_temperature_from_mixing_ratio(T,w)
    return p*100./ Rd/Tv

def pseudoadiabat(T,p):
    """Compute the local slope of the pseudoadiabat at given temperature and pressure

    Inputs:   p is pressure in hPa or mb
              T is local temperature in Kelvin
    Output:   dT/dp, the rate of temperature change for pseudoadiabatic ascent
                in units of K / hPa

    The pseudoadiabat describes changes in temperature and pressure for an air
    parcel at saturation assuming instantaneous rain-out of the super-saturated water

    Formula consistent with eq. (2.33) from Raymond Pierrehumbert, "Principles of Planetary Climate"
    which nominally accounts for non-dilute effects, but computes the derivative
    dT/dpa, where pa is the partial pressure of the non-condensible gas.

    Integrating the result dT/dp treating p as total pressure effectively makes the dilute assumption.
    """
    esoverp = clausius_clapeyron(T) / p
    Tcel = T - tempCtoK
    L = (2.501 - 0.00237 * Tcel) * 1.E6   # Accurate form of latent heat of vaporization in J/kg
    ratio = L / T /  Rv
    dTdp = (T / p * kappa * (1 + esoverp * ratio) /
        (1 + kappa * (cpv / Rv + (ratio-1) * ratio) * esoverp))
    return dTdp

def lifting_condensation_level(T, RH):
    '''Compute the Lifiting Condensation Level (LCL) for a given temperature and relative humidity

    Inputs:  T is temperature in Kelvin
            RH is relative humidity (dimensionless)

    Output: LCL in meters

    This is height (relative to parcel height) at which the parcel would become saturated during adiabatic ascent.

    Based on approximate formula from Bolton (1980 MWR) as given by Romps (2017 JAS)

    For an exact formula see Romps (2017 JAS), doi:10.1175/JAS-D-17-0102.1
    '''
    Tadj = T-55.  # in Kelvin
    return cp/g*(Tadj - (1/Tadj - log(RH)/2840.)**(-1))

def estimated_inversion_strength(T0,T700):
    '''Compute the Estimated Inversion Strength or EIS,
    following Wood and Bretherton (2006, J. Climate)

    Inputs: T0 is surface temp in Kelvin
           T700 is air temperature at 700 hPa in Kelvin

    Output: EIS in Kelvin

    EIS is a normalized measure of lower tropospheric stability acccounting for
    temperature-dependence of the moist adiabat.
    '''
    # Interpolate to 850 hPa
    T850 = (T0+T700)/2.;
    # Assume 80% relative humidity to compute LCL, appropriate for marine boundary layer
    LCL = lifting_condensation_level(T0, 0.8)
    # Lower Tropospheric Stability (theta700 - theta0)
    LTS = potential_temperature(T700, 700) - T0
    #  Gammam  = -dtheta/dz is the rate of potential temperature decrease along the moist adiabat
    #  in K / m
    Gammam = (g/cp*(1.0 - (1.0 + Lhvap*qsat(T850,850) / Rd / T850) /
             (1.0 + Lhvap**2 * qsat(T850,850)/ cp/Rv/T850**2)))
    #  Assume exponential decrease of pressure with scale height given by surface temperature
    z700 = (Rd*T0/g)*log(1000./700.)
    return LTS - Gammam*(z700 - LCL)

def EIS(T0,T700):
    '''Convenience method, identical to thermo.estimated_inversion_strength(T0,T700)'''
    return estimated_inversion_strength(T0,T700)

def blackbody_emission(T):
    '''Blackbody radiation following the Stefan-Boltzmann law.'''
    return sigma * T**4

def Planck_frequency(nu, T):
    '''The Planck function B(nu,T):
    the flux density for blackbody radiation in frequency space
    nu is frequency in 1/s
    T is temperature in Kelvin

    Formula (3.1) from Raymond Pierrehumbert, "Principles of Planetary Climate"
    '''
    return 2*hPlanck*nu**3/c_light**2/(exp(hPlanck*nu/kBoltzmann/T)-1)

def Planck_wavenumber(n, T):
    '''The Planck function (flux density for blackbody radition)
    in wavenumber space
    n is wavenumber in 1/cm
    T is temperature in Kelvin

    Formula from Raymond Pierrehumbert, "Principles of Planetary Climate", page 140.
    '''
    # convert to mks units
    n = n*100.
    return c_light * Planck_frequency(n*c_light, T)

def Planck_wavelength(l, T):
    '''The Planck function (flux density for blackbody radiation)
    in wavelength space
    l is wavelength in meters
    T is temperature in Kelvin

    Formula (3.3) from Raymond Pierrehumbert, "Principles of Planetary Climate"
    '''
    u = hPlanck*c_light/l/kBoltzmann/T
    return 2*kBoltzmann**5*T**5/hPlanck**4/c_light**3*u**5/(exp(u)-1)

def vmr_to_mmr(vmr, gas):
    '''
    Convert volume mixing ratio to mass mixing ratio for named gas.
    ( molecular weights are specific in climlab.utils.constants.py )
    '''
    return vmr * molecular_weight[gas] / molecular_weight['dry air']

def mmr_to_vmr(mmr, gas):
    '''
    Convert mass mixing ratio to volume mixing ratio for named gas.
    ( molecular weights are specific in climlab.utils.constants.py )
    '''
    return mmr * molecular_weight['dry air'] / molecular_weight[gas]



from __future__ import division
from climlab.process.energy_budget import EnergyBudget
from climlab.utils import constants as const
import numpy as np


class AplusBT(EnergyBudget):
    r"""The simplest linear longwave radiation module.

    Calculates the Outgoing Longwave Radation (OLR) :math:`R\uparrow` as

    .. math::

        R\uparrow = A + B \cdot T

    where :math:`T` is the state variable.

    Should be invoked with a single temperature state variable only.


    **Initialization parameters** \n

    An instance of ``AplusBT`` is initialized with the following
    arguments:

    :param float A:             parameter for linear OLR parametrization   \n
                                - unit: :math:`\frac{\textrm{W}}{\textrm{m}^2}` \n
                                - default value: ``200.0``
    :param float B:             parameter for linear OLR parametrization   \n
                                - unit: :math:`\frac{\textrm{W}} {\textrm{m}^2 ^{\circ}\textrm{C}}`  \n
                                - default value: ``2.0``

    **Object attributes** \n

    Additional to the parent class :class:`~climlab.process.energy_budget.EnergyBudget`
    following object attributes are generated or modified during initialization:

    :ivar float A:              calls the setter function of :func:`A`
    :ivar float B:              calls the setter function of :func:`B`
    :ivar dict diagnostics:     key ``'OLR'`` initialized with value:
                                :class:`~climlab.domain.field.Field` of zeros
                                in size of ``self.Ts``
    :ivar Field OLR:            the subprocess attribute ``self.OLR`` is
                                created with correct dimensions


    .. warning::

        This module currently works only for a single state variable!

    :Example:

        Simple linear radiation module (stand alone)::

            >>> import climlab

            >>> # create a column atmosphere and scalar surface
            >>> sfc, atm = climlab.domain.single_column()

            >>> # Create a state variable
            >>> Ts = climlab.Field(15., domain=sfc)

            >>> # Make a dictionary of state variables
            >>> s = {'Ts': Ts}

            >>> # create process
            >>> olr = climlab.radiation.AplusBT(state=s)

            >>> print olr
            climlab Process of type <class 'climlab.radiation.AplusBT.AplusBT'>.
            State variables and domain shapes:
              Ts: (1,)
            The subprocess tree:
            top: <class 'climlab.radiation.AplusBT.AplusBT'>

            >>> # to compute tendencies and diagnostics
            >>> olr.compute()

            >>> #  or to actually update the temperature
            >>> olr.step_forward()

            >>> print olr.state
            {'Ts': Field([ 5.69123176])}

    """
    def __init__(self, A=200., B=2., **kwargs):
        super(AplusBT, self).__init__(**kwargs)
        self.A = A
        self.B = B
        self.add_diagnostic('OLR', 0. * self.Ts)

    @property
    def A(self):
        """Property of AplusBT parameter A.

        :getter:    Returns the parameter A which is stored in attribute
                    ``self._A``
        :setter:    * sets parameter A which is addressed as ``self._A``
                      to the new value
                    * updates the parameter dictionary ``self.param['A']``
        :type:      float

        :Example:

            ::

                >>> import climlab
                >>> model = climlab.EBM()

                >>> # getter
                >>> model.subprocess['LW'].A
                210.0
                >>> # setter
                >>> model.subprocess['LW'].A = 220
                >>> # getter again
                >>> model.subprocess['LW'].A
                220

                >>> # subprocess parameter dictionary
                >>> model.subprocess['LW'].param['A']
                220

        """
        return self._A
    @A.setter
    def A(self, value):
        self._A = value
        self.param['A'] = value
    @property
    def B(self):
        """Property of AplusBT parameter B.

        :getter:    Returns the parameter B which is stored in attribute
                    ``self._B``
        :setter:    * sets parameter B which is addressed as ``self._B``
                      to the new value
                    * updates the parameter dictionary ``self.param['B']``
        :type:      float

        """
        return self._B
    @B.setter
    def B(self, value):
        self._B = value
        self.param['B'] = value

    def _compute_emission(self):
        for varname, value in self.state.items():
            self.OLR[:] = self.A + self.B * value

    def _compute_heating_rates(self):
        '''Compute energy flux convergences to get heating rates in :math:`W/m^2`,'''
        self._compute_emission()
        for varname, value in self.state.items():
            self.heating_rate[varname] = -self.OLR


class AplusBT_CO2(EnergyBudget):
    """Linear longwave radiation module considering CO2 concentration.

    This radiation subprocess is based in the idea to linearize the Outgoing
    Longwave Radiation (OLR) emitted to space according to the surface temperature
    (see :class:`AplusBT`).

    To consider a the change of the greenhouse effect through range of
    :math:`CO_2` in the atmosphere, the parameters A and B are computed like
    the following:

    .. math::

        A(c) = -326.4 + 9.161 c - 3.164 c^2 + 0.5468 c^3            \n
        B(c) =  1.953 - 0.04866 c + 0.01309 c^2 - 0.002577 c^3

    where :math:`c=\\log \\frac{p}{300}` and :math:`p` represents
    the concentration of :math:`CO_2` in the atmosphere.

    For further reading see :cite:`Caldeira_1992`.


    **Initialization parameters** \n

    An instance of ``AplusBT_CO2`` is initialized with the following
    argument:

    :param float CO2:   The concentration of :math:`CO_2` in the atmosphere.
                        Referred to as :math:`p` in the above given formulas.\n
                        - unit: :math:`\\textrm{ppm}` (parts per million)   \n
                        - default value: ``300.0``


    **Object attributes** \n

    Additional to the parent class :class:`~climlab.process.energy_budget.EnergyBudget`
    following object attributes are generated or updated during initialization:

    :ivar float CO2:                calls the setter function of :func:`CO2`
    :ivar dict diagnostics:         the subprocess's diagnostic dictionary
                                    ``self.diagnostic`` is initialized
                                    through calling
                                    ``self.add_diagnostic('OLR', 0. * self.Ts)``
    :ivar Field OLR:                the subprocess attribute ``self.OLR`` is
                                    created with correct dimensions

    :Example:

        Replacing an the regular AplusBT subprocess in an energy balance model::

            >>> import climlab
            >>> from climlab.radiation.AplusBT import AplusBT_CO2

            >>> # creating EBM model
            >>> model = climlab.EBM()

            >>> print model

        .. code-block:: none
            :emphasize-lines: 7

            climlab Process of type <class 'climlab.model.ebm.EBM'>.
            State variables and domain shapes:
              Ts: (90, 1)
            The subprocess tree:
            top: <class 'climlab.model.ebm.EBM'>
               diffusion: <class 'climlab.dynamics.diffusion.MeridionalDiffusion'>
               LW: <class 'climlab.radiation.AplusBT.AplusBT'>
               albedo: <class 'climlab.surface.albedo.StepFunctionAlbedo'>
                  iceline: <class 'climlab.surface.albedo.Iceline'>
                  cold_albedo: <class 'climlab.surface.albedo.ConstantAlbedo'>
                  warm_albedo: <class 'climlab.surface.albedo.P2Albedo'>
               insolation: <class 'climlab.radiation.insolation.P2Insolation'>

        ::

            >>> #  creating and adding albedo feedback subprocess
            >>> LW_CO2 = AplusBT_CO2(CO2=400, state=model.state, **model.param)

            >>> # overwriting old 'LW' subprocess with same name
            >>> model.add_subprocess('LW', LW_CO2)

            >>> print model

        .. code-block:: none
            :emphasize-lines: 7

            climlab Process of type <class 'climlab.model.ebm.EBM'>.
            State variables and domain shapes:
              Ts: (90, 1)
            The subprocess tree:
            top: <class 'climlab.model.ebm.EBM'>
               diffusion: <class 'climlab.dynamics.diffusion.MeridionalDiffusion'>
               LW: <class 'climlab.radiation.AplusBT.AplusBT_CO2'>
               albedo: <class 'climlab.surface.albedo.StepFunctionAlbedo'>
                  iceline: <class 'climlab.surface.albedo.Iceline'>
                  cold_albedo: <class 'climlab.surface.albedo.ConstantAlbedo'>
                  warm_albedo: <class 'climlab.surface.albedo.P2Albedo'>
               insolation: <class 'climlab.radiation.insolation.P2Insolation'>

    """
    # implemented by m-kreuzer
    def __init__(self, CO2=300., **kwargs):
        super(AplusBT_CO2, self).__init__(**kwargs)
        self.CO2 = CO2
        #newdiags = ['OLR',]
        #self.add_diagnostics(newdiags)
        self.add_diagnostic('OLR', 0. * self.Ts)

    @property
    def CO2(self):
        """Property of AplusBT_CO2 parameter CO2.

        :getter:    Returns the CO2 concentration which is stored in attribute
                    ``self._CO2``
        :setter:    * sets the CO2 concentration which is addressed as ``self._CO2``
                      to the new value
                    * updates the parameter dictionary ``self.param['CO2']``
        :type:      float

        """
        return self._CO2
    @CO2.setter
    def CO2(self, value):
        self._CO2 = value
        self.param['CO2'] = value

#    def emission(self):
#        """Calculates the Outgoing Longwave Radiation (OLR) of the AplusBT_CO2
#        subprocess.
#
#        **Object attributes** \n
#
#        During method execution following object attribute is modified:
#
#        :ivar float OLR:            the described formula is calculated and the
#                                    result stored in the project attribute ``self.OLR``
#        :ivar dict diagnostics:     the same result is written in ``diagnostics``
#                                    dictionary with the key ``'OLR'``
#
#        .. warning::
#
#            This method currently works only for a single state variable!
#
#        """
#        l = np.log(self.CO2/300.)
#        A = -326.400 + 9.16100*l - 3.16400*l**2 + 0.546800*l**3
#        B =    1.953 - 0.04866*l + 0.01309*l**2 - 0.002577*l**3
#        for varname, value in self.state.iteritems():
#            flux = A + B * (value + const.tempCtoK)
#            self.OLR = flux
#            self.diagnostics['OLR'] = self.OLR

    def _compute_emission(self):
        l = np.log(self.CO2/300.)
        self.A = -326.400 + 9.16100*l - 3.16400*l**2 + 0.546800*l**3
        self.B =    1.953 - 0.04866*l + 0.01309*l**2 - 0.002577*l**3
        for varname, value in self.state.items():
            self.OLR[:] = self.A + self.B * (value + const.tempCtoK)

    def _compute_heating_rates(self):
        """Computes energy flux convergences to get heating rates in :math:`W/m^2`."""
        self._compute_emission()
        for varname, value in self.state.items():
            self.heating_rate[varname] = -self.OLR



from __future__ import division
from builtins import range
import numpy as np
from climlab import constants as const
from climlab.utils.thermo import rho_moist, pseudoadiabat
from climlab.process.time_dependent_process import TimeDependentProcess
from climlab.domain.field import Field
from .akmaev_adjustment import convective_adjustment_direct


class ConvectiveAdjustment(TimeDependentProcess):
    '''Hard Convective Adjustment to a prescribed lapse rate.

    This process computes the instantaneous adjustment to conservatively
    remove any instabilities in each column.

    Instability is defined as a temperature decrease with height that exceeds
    the prescribed critical lapse rate. This critical rate is set by input argument
    ``adj_lapse_rate``, which can be either a numerical or string value.

    Numerical values for ``adj_lapse_rate`` are given in units of K / km. Both
    array and scalar values are valid. For scalar values, the assumption is that
    the critical lapse rate is the same at every level.

    If an array is given, it is assumed to represent the in-situ critical lapse
    rate (in K/km) at every grid point.

    Alternatively, string arguments can be given as follows:

        - ``'DALR'`` or ``'dry adiabat'``: critical lapse rate is set to g/cp = 9.8 K / km
        - ``'MALR'`` or ``'moist adiabat'`` or ``'pseudoadiabat'``: critical lapse rate follows the in-situ moist pseudoadiabat at every level

    Adjustment includes the surface if ``'Ts'`` is included in the ``state``
    dictionary. This implicitly accounts for turbulent surface fluxes.
    Otherwise only the atmospheric temperature is adjusted.

    If ``adj_lapse_rate`` is an array, its size must match the number of vertical
    levels of the adjustment. This is number of pressure levels if the surface is
    not adjusted, or number of pressure levels + 1 if the surface is adjusted.

    This process implements the conservative adjustment algorithm described in
    Akmaev (1991) Monthly Weather Review.
    '''
    def __init__(self, adj_lapse_rate=None, **kwargs):
        super(ConvectiveAdjustment, self).__init__(**kwargs)
        # lapse rate for convective adjustment, in K / km
        self.adj_lapse_rate = adj_lapse_rate
        self.param['adj_lapse_rate'] = adj_lapse_rate
        self.time_type = 'adjustment'
        self.adjustment = {}
    @property
    def pcol(self):
        patm = self.lev
        if 'Ts' in self.state:
            #  surface pressure should correspond to model domain!
            ps = self.lev_bounds[-1]
            return np.append(patm, ps)
        else:
            return patm
    @property
    def ccol(self):
        c_atm = self.Tatm.domain.heat_capacity
        if 'Ts' in self.state:
            c_sfc = self.Ts.domain.heat_capacity
            return np.append(c_atm, c_sfc)
        else:
            return c_atm
    @property
    def Tcol(self):
        #  For now, let's assume that the vertical axis is the last axis
        Tatm = self.Tatm
        if 'Ts' in self.state:
            Ts = np.atleast_1d(self.Ts)
            return np.concatenate((Tatm, Ts),axis=-1)
        else:
            return Tatm
    @property
    def adj_lapse_rate(self):
        lapserate = self._adj_lapse_rate
        if type(lapserate) is str:
            if lapserate in ['DALR', 'dry adiabat']:
                return const.g / const.cp * 1.E3
            elif lapserate in ['MALR', 'moist adiabat', 'pseudoadiabat']:
                # critical lapse rate at each level is set by pseudoadiabat
                dTdp = pseudoadiabat(self.Tcol,self.pcol) / 100.  # K / Pa
                #  Could include water vapor effect on density here ...
                #  Replace Tcol with virtual temperature
                rho = self.pcol*100./const.Rd/self.Tcol  # in kg/m**3
                return dTdp * const.g * rho * 1000.  # K / km
            else:
                raise ValueError('adj_lapse_rate must be either numeric or any of \'DALR\', \'dry adiabat\', \'MALR\', \'moist adiabat\', \'pseudoadiabat\'.')
        else:
            return lapserate
    @adj_lapse_rate.setter
    def adj_lapse_rate(self, lapserate):
        self._adj_lapse_rate = lapserate
        self.param['adj_lapse_rate'] = lapserate

    def _compute(self):
        if self.adj_lapse_rate is None:
            self.adjustment['Ts'] = self.Ts * 0.
            self.adjustment['Tatm'] = self.Tatm * 0.
        else:
            #  convective adjustment routine expect reversered vertical axis
            pflip = self.pcol[..., ::-1]
            Tflip = self.Tcol[..., ::-1]
            cflip = self.ccol[..., ::-1]
            lapseflip = np.atleast_1d(self.adj_lapse_rate)[..., ::-1]
            Tadj_flip = convective_adjustment_direct(pflip, Tflip, cflip, lapserate=lapseflip)
            Tadj = Tadj_flip[..., ::-1]
            if 'Ts' in self.state:
                Ts = Field(Tadj[...,-1], domain=self.Ts.domain)
                Tatm = Field(Tadj[...,:-1], domain=self.Tatm.domain)
                self.adjustment['Ts'] = Ts - self.Ts
            else:
                Tatm = Field(Tadj, domain=self.Tatm.domain)
            self.adjustment['Tatm'] = Tatm - self.Tatm
        #  return the adjustment, independent of timestep
        #  because the parent process might have set a different timestep!
        return self.adjustment



from __future__ import division
import numpy as np
from climlab.process.time_dependent_process import TimeDependentProcess


class EnergyBudget(TimeDependentProcess):
    r"""A parent class for explicit energy budget processes.

    This class solves equations that include a heat capacitiy term like
    :math:`C \\frac{dT}{dt} = \\textrm{flux convergence}`

    In an Energy Balance Model with model state :math:`T` this equation
    will look like this:

    .. math::

        C \\frac{dT}{dt} = R\downarrow - R\uparrow - H  \n
        \\frac{dT}{dt} = \\frac{R\downarrow}{C} - \\frac{R\uparrow}{C} - \\frac{H}{C}

    Every EnergyBudget object has a ``heating_rate`` dictionary with items
    corresponding to each state variable. The heating rate accounts the actual
    heating of a subprocess, namely the contribution to the energy budget
    of :math:`R\\downarrow, R\\uparrow` and :math:`H` in this case.
    The temperature tendencies for each subprocess are then calculated
    through dividing the heating rate by the heat capacitiy :math:`C`.

    **Initialization parameters** \n

    An instance of ``EnergyBudget`` is initialized with the forwarded
    keyword arguments ``**kwargs`` of the corresponding children classes.

    **Object attributes** \n

    Additional to the parent class
    :class:`~climlab.process.timedependentprocess.TimeDependentProcess`
    following object attributes are generated or modified during initialization:

    :ivar str time_type:        is set to ``'explicit'``
    :ivar dict heating_rate:    energy share for given subprocess in unit
                                :math:`\\textrm{W}/ \\textrm{m}^2` stored
                                in a dictionary sorted by model states

    """
    def __init__(self, **kwargs):
        super(EnergyBudget, self).__init__(**kwargs)
        self.time_type = 'explicit'
        self.heating_rate = {}

    def _compute_heating_rates(self):
        """Computes energy flux convergences to get heating rates in unit
        :math:`\\textrm{W}/ \\textrm{m}^2`.

        This method should be over-ridden by daughter classes.

        """
        for varname in list(self.state.keys()):
            self.heating_rate[varname] = self.state[varname] * 0.

    def _temperature_tendencies(self):
        self._compute_heating_rates()
        tendencies = {}
        for varname, value in self.state.items():
            #C = self.state_domain[varname].heat_capacity
            C = value.domain.heat_capacity
            try:  # there may be state variables without heating rates
                tendencies[varname] = (self.heating_rate[varname] / C)
            except:
                pass
        return tendencies

    def _compute(self):
        tendencies = self._temperature_tendencies()
        return tendencies


class ExternalEnergySource(EnergyBudget):
    """A fixed energy source or sink to be specified by the user.

    **Object attributes** \n

    Additional to the parent class :class:`~climlab.process.energy_budget.EnergyBudget`
    the following object attribute is modified during initialization:

    :ivar dict heating_rate:    energy share dictionary for this subprocess
                                is set to zero for every model state.

    After initialization the user should modify the fields in the
    ``heating_rate`` dictionary, which contain heating rates in
    unit :math:`\\textrm{W}/ \\textrm{m}^2` for all state variables.

    :Example:

        Creating an Energy Balance Model with a uniform external energy source
        of :math:`10 \\ \\textrm{W}/ \\textrm{m}^2` for all latitudes::

            >>> import climlab
            >>> from climlab.process.energy_budget import ExternalEnergySource
            >>> import numpy as np

            >>> # create model & external energy subprocess
            >>> model = climlab.EBM(num_lat=36)
            >>> ext_en = ExternalEnergySource(state= model.state,**model.param)

            >>> # modify external energy rate
            >>> ext_en.heating_rate.keys()
            ['Ts']

            >>> np.squeeze(ext_en.heating_rate['Ts'])
            Field([-0., -0., -0., -0., -0., -0., -0., -0., -0.,  0.,  0.,  0.,  0.,
                    0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,
                    0., -0., -0., -0., -0., -0., -0., -0., -0., -0.])

            >>> ext_en.heating_rate['Ts'][:]=10

            >>> np.squeeze(ext_en.heating_rate['Ts'])
            Field([ 10.,  10.,  10.,  10.,  10.,  10.,  10.,  10.,  10.,  10.,  10.,
                    10.,  10.,  10.,  10.,  10.,  10.,  10.,  10.,  10.,  10.,  10.,
                    10.,  10.,  10.,  10.,  10.,  10.,  10.,  10.,  10.,  10.,  10.,
                    10.,  10.,  10.])

            >>> # add subprocess to model
            >>> model.add_subprocess('ext_energy',ext_en)

            >>> print model
            climlab Process of type <class 'climlab.model.ebm.EBM'>.
            State variables and domain shapes:
              Ts: (36, 1)
            The subprocess tree:
            top: <class 'climlab.model.ebm.EBM'>
               diffusion: <class 'climlab.dynamics.diffusion.MeridionalDiffusion'>
               LW: <class 'climlab.radiation.AplusBT.AplusBT'>
               ext_energy: <class 'climlab.process.energy_budget.ExternalEnergySource'>
               albedo: <class 'climlab.surface.albedo.StepFunctionAlbedo'>
                  iceline: <class 'climlab.surface.albedo.Iceline'>
                  cold_albedo: <class 'climlab.surface.albedo.ConstantAlbedo'>
                  warm_albedo: <class 'climlab.surface.albedo.P2Albedo'>
               insolation: <class 'climlab.radiation.insolation.P2Insolation'>

    """
    def __init__(self, **kwargs):
        super(ExternalEnergySource, self).__init__(**kwargs)
        for varname in list(self.state.keys()):
            self.heating_rate[varname] = self.state[varname] * 0.

    def _compute_heating_rates(self):
        pass



import climlab
import matplotlib.pyplot as plt

# creating & integrating model
model = climlab.EBM()
model.step_forward()

# plot
fig = plt.figure( figsize=(6,4))
ax = fig.add_subplot(111)

bounds = model.domains['Ts'].axes['lat'].bounds
ax.plot(bounds, model.heat_transport())

ax.set_title('heat transport')
ax.set_xlabel('latitude')
ax.set_xticks([-90,-60,-30,0,30,60,90])
ax.set_ylabel('energy (PW)')
plt.axhline(linewidth=2, color='grey', linestyle='dashed')
plt.show()



'''Modules for surface processes in climlab.

``climlab.surface.turbulent`` contains ``SensibleHeatFlux`` and ``LatentHeatFlux``
processes that implement bulk aerodynamic fluxes for surface energy and water exchange.

``climlab.surface.albedo`` contains processes for surface albedo
and interactive ice and snow lines for energy balance models.
'''
from __future__ import absolute_import
from .turbulent import SensibleHeatFlux, LatentHeatFlux
from .albedo import ConstantAlbedo, P2Albedo, Iceline, StepFunctionAlbedo



'''
A climlab process for the Emanuel convection scheme
'''
from __future__ import absolute_import

import numpy as np
import warnings
from climlab.process import TimeDependentProcess
from climlab.utils.thermo import qsat
from climlab import constants as const
try:
    from climlab_emanuel_convection import emanuel_convection as convect
except:
    warnings.warn('Cannot import EmanuelConvection fortran extension, this module will not be functional.')
# The array conversion routines we are borrowing from the RRTMG wrapper
from climlab.radiation.rrtm.utils import _climlab_to_rrtm as _climlab_to_convect
from climlab.radiation.rrtm.utils import _rrtm_to_climlab as _convect_to_climlab


# Thermodynamic constants
CPD = const.cp
CPV = const.cpv
RV = const.Rv
RD = const.Rd
LV0 = const.Lhvap
G = const.g
ROWL = const.rho_w
# specific heat of liquid water -- artifically small!
#   Kerry Emanuel's notes say this is intentional, do not change this.
CL=2500.0
#CPV = CPD  #  try neglecting effect of water vapor on heat capacity

class EmanuelConvection(TimeDependentProcess):
    '''
    The climlab wrapper for Kerry Emanuel's moist convection scheme <https://emanuel.mit.edu/FORTRAN-subroutine-convect>

    From the documentation distributed with the Fortran 77 code CONVECT:

        The subroutine is designed to be used in time-marching models of mesoscale to global-scale dimensions.
        It is meant to represent the effects of all moist convection, including shallow, non-precipitating cumulus.
        It also contains a dry adiabatic adjustment scheme.

        Since the method of calculating the convective fluxes involves a relaxation toward quasi-equilibrium,
        subroutine CONVECT must be run for at least several time steps to give meaningful results.
        At the first time step, the tendencies and convective precipitation will be zero.
        If the initial sounding is unstable, these will rapidly increase over successive time steps,
        depending on the values of the constants ALPHA and DAMP.
        Thus the user interested in convective fluxes and precipitation
        associated with a single initial sounding (i.e., without large-scale forcing)
        should still march CONVECT forward enough time steps that the fluxes have
        returned back to zero;
        the net tendencies and precipitation integrated over this time interval are then the desired results.
        But it should be cautioned that these quantities will not necessarily be
        independent of other model parameters such as the time step.
        CONVECT is very much built on the philosophy that convection,
        to the extent it can be represented in terms of large-scale variables,
        is never very far away from statistical equilibrium with the large-scale flow.
        To achieve a smooth evolution of the convective forcing,
        CONVECT should be called at least every 20 minutes during the time integration.
        CONVECT will work at longer time intervals, but the convective tendencies may become noisy.

    Basic characteristics:

    State:

        - Ts (surface radiative temperature -- optional, and ignored)
        - Tatm (air temperature in K)
        - q (specific humidity in kg/kg)
        - U (zonal velocity in m/s -- optional)
        - V (meridional velocity in m/s -- optional)

    Input arguments and default values (taken from convect43.f fortran source):

        - MINORIG = 0,  index of lowest level from which convection may originate (zero means lowest)
        - ELCRIT = 0.0011,  autoconversion threshold water content (g/g)
        - TLCRIT = -55.0, critical temperature below which the auto-conversion threshold is assumed to be zero (the autoconversion threshold varies linearly between 0 C and TLCRIT)
        - ENTP = 1.5, coefficient of mixing in the entrainment formulation
        - SIGD = 0.05, fractional area covered by unsaturated downdraft
        - SIGS = 0.12, fraction of precipitation falling outside of cloud
        - OMTRAIN = 50.0, assumed fall speed (Pa/s) of rain
        - OMTSNOW = 5.5, assumed fall speed (Pa/s) of snow
        - COEFFR = 1.0, coefficient governing the rate of evaporation of rain
        - COEFFS = 0.8, coefficient governing the rate of evaporation of snow
        - CU = 0.7, coefficient governing convective momentum transport
        - BETA = 10.0, coefficient used in downdraft velocity scale calculation
        - DTMAX = 0.9, maximum negative temperature perturbation a lifted parcel is allowed to have below its LFC
        - ALPHA = 0.2, first parameter that controls the rate of approach to quasi-equilibrium
        - DAMP = 0.1, second parameter that controls the rate of approach to quasi-equilibrium (DAMP must be less than 1)
        - IPBL = 0, switch to bypass the dry convective adjustment (bypass if IPBL==0)

    Tendencies computed:

        - air temperature (K/s)
        - specific humidity (kg/kg/s)
        - optional:
            - U and V wind components (m/s/s), if U and V are included in state dictionary

    Diagnostics computed:

        - CBMF (cloud base mass flux in kg/m2/s) -- this is actually stored internally and used as input for subsequent timesteps
        - precipitation (convective precipitation rate in kg/m2/s or mm/s)
        - relative_humidity (dimensionless)

        :Example:

            Here is an example of setting up a single-column
            Radiative-Convective model with interactive water vapor.

            This example also demonstrates *asynchronous coupling*:
            the radiation uses a longer timestep than the other model components::

                import numpy as np
                import climlab
                from climlab import constants as const
                # Temperatures in a single column
                full_state = climlab.column_state(num_lev=30, water_depth=2.5)
                temperature_state = {'Tatm':full_state.Tatm,'Ts':full_state.Ts}
                #  Initialize a nearly dry column (small background stratospheric humidity)
                q = np.ones_like(full_state.Tatm) * 5.E-6
                #  Add specific_humidity to the state dictionary
                full_state['q'] = q
                #  ASYNCHRONOUS COUPLING -- the radiation uses a much longer timestep
                #  The top-level model
                model = climlab.TimeDependentProcess(state=full_state,
                                              timestep=const.seconds_per_hour)
                #  Radiation coupled to water vapor
                rad = climlab.radiation.RRTMG(state=temperature_state,
                                              specific_humidity=full_state.q,
                                              albedo=0.3,
                                              timestep=const.seconds_per_day
                                              )
                #  Convection scheme -- water vapor is a state variable
                conv = climlab.convection.EmanuelConvection(state=full_state,
                                              timestep=const.seconds_per_hour)
                #  Surface heat flux processes
                shf = climlab.surface.SensibleHeatFlux(state=temperature_state, Cd=0.5E-3,
                                              timestep=const.seconds_per_hour)
                lhf = climlab.surface.LatentHeatFlux(state=full_state, Cd=0.5E-3,
                                              timestep=const.seconds_per_hour)
                #  Couple all the submodels together
                model.add_subprocess('Radiation', rad)
                model.add_subprocess('Convection', conv)
                model.add_subprocess('SHF', shf)
                model.add_subprocess('LHF', lhf)
                print(model)

                #  Run the model
                model.integrate_years(1)
                #  Check for energy balance
                print(model.ASR - model.OLR)
    '''
    def __init__(self,
            MINORIG = 0,  # index of lowest level from which convection may originate (zero means lowest)
            #  Default parameter values taken from convect43c.f  fortran source
            ELCRIT=.0011,
            TLCRIT=-55.0,
            ENTP=1.5,
            SIGD=0.05,
            SIGS=0.12,
            OMTRAIN=50.0,
            OMTSNOW=5.5,
            COEFFR=1.0,
            COEFFS=0.8,
            CU=0.7,
            BETA=10.0,
            DTMAX=0.9,
            ALPHA=0.2,
            DAMP=0.1,
            IPBL=0,
            **kwargs):
        super(EmanuelConvection, self).__init__(**kwargs)
        self.time_type = 'explicit'
        #  Define inputs and diagnostics
        surface_shape = self.state['Tatm'][...,0].shape
        #  Hack to handle single column and multicolumn
        if surface_shape == ():
            init = np.atleast_1d(np.zeros(surface_shape))
            self.multidim=False
        else:
            init = np.zeros(surface_shape)[...,np.newaxis]
            self.multidim=True
        self.add_diagnostic('CBMF', init*0.)  # cloud base mass flux
        self.add_diagnostic('precipitation', init*0.) # Precip rate (kg/m2/s)
        self.add_diagnostic('relative_humidity', 0*self.Tatm)
        self.add_input('MINORIG', MINORIG)
        self.add_input('ELCRIT', ELCRIT)
        self.add_input('TLCRIT', TLCRIT)
        self.add_input('ENTP', ENTP)
        self.add_input('SIGD', SIGD)
        self.add_input('SIGS', SIGS)
        self.add_input('OMTRAIN', OMTRAIN)
        self.add_input('OMTSNOW', OMTSNOW)
        self.add_input('COEFFR', COEFFR)
        self.add_input('COEFFS', COEFFS)
        self.add_input('CU', CU)
        self.add_input('BETA', BETA)
        self.add_input('DTMAX', DTMAX)
        self.add_input('ALPHA', ALPHA)
        self.add_input('DAMP', DAMP)
        self.add_input('IPBL', IPBL)

    def _compute(self):
        #  Invert arrays so the first element is the bottom of column
        T = _climlab_to_convect(self.state['Tatm'])
        dom = self.state['Tatm'].domain
        P = _climlab_to_convect(dom.lev.points)
        PH = _climlab_to_convect(dom.lev.bounds)
        Q = _climlab_to_convect(self.state['q'])
        QS = qsat(T,P)
        ND = np.size(T, axis=1)
        NCOL = np.size(T, axis=0)
        NL = ND-1
        try:
            U = _climlab_to_convect(self.state['U'])
        except:
            U = np.zeros_like(T)
        try:
            V = _climlab_to_convect(self.state['V'])
        except:
            V = np.zeros_like(T)
        NTRA = 1
        TRA = np.zeros((NCOL,ND,NTRA), order='F')  # tracers ignored
        DELT = float(self.timestep)
        CBMF = self.CBMF
        (IFLAG, FT, FQ, FU, FV, FTRA, PRECIP, WD, TPRIME, QPRIME, CBMFnew,
        Tout, Qout, QSout, Uout, Vout, TRAout) = \
            convect(T, Q, QS, U, V, TRA, P, PH, NCOL, ND, NL, NTRA, DELT, self.IPBL, CBMF,
                    CPD, CPV, CL, RV, RD, LV0, G, ROWL, self.MINORIG,
                    self.ELCRIT, self.TLCRIT, self.ENTP, self.SIGD, self.SIGS,
                    self.OMTRAIN, self.OMTSNOW, self.COEFFR, self.COEFFS,
                    self.CU, self.BETA, self.DTMAX, self.ALPHA, self.DAMP
                    )
        # If dry adjustment is being used then the tendencies need to be adjusted
        if self.IPBL != 0:
            FT += (Tout - T) / DELT
            FQ += (Qout - Q) / DELT
        tendencies = {'Tatm': _convect_to_climlab(FT)*np.ones_like(self.state['Tatm']),
                      'q': _convect_to_climlab(FQ)*np.ones_like(self.state['q'])}
        if 'Ts' in self.state:
            # for some strange reason self.Ts is breaking tests under Python 3.5 in some configurations
            tendencies['Ts'] = 0. * self.state['Ts']
        if 'U' in self.state:
            tendencies['U'] = _convect_to_climlab(FU) * np.ones_like(self.state['U'])
        if 'V' in self.state:
            tendencies['V'] = _convect_to_climlab(FV) * np.ones_like(self.state['V'])
        self.CBMF = CBMFnew
        #  Need to convert from mm/day to mm/s or kg/m2/s
        #  Hack to handle single column and multicolumn
        if self.multidim:
            self.precipitation[:,0] = _convect_to_climlab(PRECIP)/const.seconds_per_day
        else:
            self.precipitation[:] = _convect_to_climlab(PRECIP)/const.seconds_per_day
        self.IFLAG = IFLAG
        self.relative_humidity[:] = self.q / qsat(self.Tatm,self.lev)
        return tendencies



from __future__ import division, print_function
from climlab.process.energy_budget import EnergyBudget
from climlab import constants as const


class SimpleAbsorbedShortwave(EnergyBudget):
    '''A class for the shortwave radiation process in a one-layer EBM.
    The basic assumption is that all the shortwave absorption occurs at the surface.

    Computes the diagnostic ``ASR`` (absorbed shortwave radiation)
    from the formula ``self.ASR = (1-self.albedo) * self.insolation``
    and applies this as a tendency on the surface temperature ``self.Ts``

    ``albedo`` and ``insolation`` are given as inputs.
    These should either be scalars
    or have same dimensions as state variable ``Ts``

    User can supply constants, or link to diagnostics of specific insolation
    and albedo processes.
    '''
    def __init__(self,
                 insolation=const.S0/4,
                 albedo=0.3,
                 **kwargs):
        super(SimpleAbsorbedShortwave, self).__init__(**kwargs)
        self.add_input('albedo', albedo)
        self.add_input('insolation', insolation)
        self.add_diagnostic('ASR', 0.*self.Ts)
        self.topdown = False  # call subprocess compute methods first

    def _compute_heating_rates(self):
        self.ASR = (1-self.albedo) * self.insolation
        self.heating_rate['Ts'] = self.ASR



from __future__ import division
import numpy as np
import climlab
from climlab.convection import emanuel_convection
from climlab.tests.xarray_test import to_xarray
import pytest
import sys


#  These test data are based on direct single-column tests of the CONVECT43c.f
#  fortran source code. We are just checking to see if we get the right tendencies
num_lev = 20
#  INPUT DATA
T = np.flipud([278.0, 273.9, 269.8, 265.7, 261.6, 257.5, 253.4, 249.3, 245.2,
    241.1, 236.9, 232.8, 228.7, 224.6, 220.5, 216.4, 212.3, 214.0, 240., 270.])
Q = np.flipud([3.768E-03, 2.812E-03, 2.078E-03, 1.519E-03, 1.099E-03,
            7.851E-04, 5.542E-04, 3.860E-04, 2.652E-04, 1.794E-04,
            1.183E-04, 7.739E-05, 4.970E-05, 3.127E-05, 1.923E-05,
            1.152E-05, 6.675E-06, 5.000E-06, 5.000E-06, 5.000E-06])
U = np.flipud([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 11.0,
                12.0, 13.0, 14.0, 15.0, 16.0, 17.0, 18.0, 19.0, 20.0])
V = 5. * np.ones_like(U)
DELT = 60.0*10.
#  TENDENCIES FROM FORTRAN CODE
FT = np.flipud(np.array([-1.79012722e-05, -5.30474102e-06, -1.31791678e-05,
        -1.52691407e-06,  2.39796248e-05,  5.00321776e-05,
         5.81085251e-05,  3.53241836e-05,  2.92659212e-05,
         1.72941819e-05, -1.29258611e-05, -1.95583598e-05,
         0.00000000e+00,  0.00000000e+00,  0.00000000e+00,
         0.00000000e+00,  0.00000000e+00,  0.00000000e+00,
         0.00000000e+00,  0.00000000e+00]))
FQ = np.flipud(np.array([-1.25265799e-07, -1.77207007e-08,  2.25623932e-08,
         1.20605250e-08, -2.24795916e-09, -8.65551558e-09,
         1.32081499e-08,  3.48952331e-08,  4.61431533e-09,
         3.59264661e-09,  3.54263054e-09,  1.12591034e-09,
         0.00000000e+00,  0.00000000e+00,  0.00000000e+00,
         0.00000000e+00,  0.00000000e+00,  0.00000000e+00,
         0.00000000e+00,  0.00000000e+00]))
FU = np.flipud(np.array([ 6.96135486e-05,  2.54272652e-05, -4.23745237e-06,
        -2.25829445e-06,  5.97676516e-06,  1.29817497e-05,
        -7.07183988e-06, -5.06040666e-05, -8.67351863e-06,
        -1.08615293e-05, -1.97420065e-05, -1.05506113e-05,
         0.00000000e+00,  0.00000000e+00,  0.00000000e+00,
         0.00000000e+00,  0.00000000e+00,  0.00000000e+00,
         0.00000000e+00,  0.00000000e+00]))
FV = np.zeros_like(FU)
CBMF = 0.0310373

#  Set thermodynamic constants to their defaults from Emanuel's code
#   so that we get same tendencies
emanuel_convection.CPD=1005.7
emanuel_convection.CPV=1870.0
emanuel_convection.RV=461.5
emanuel_convection.RD=287.04
emanuel_convection.LV0=2.501E6
emanuel_convection.G=9.8
emanuel_convection.ROWL=1000.0

#@pytest.mark.skipif(sys.platform == "darwin", reason="problematic on Mac OS for some reason")
@pytest.mark.compiled
@pytest.mark.fast
def test_convect_tendencies():
    # Temperatures in a single column
    state = climlab.column_state(num_lev=num_lev)
    state.Tatm[:] = T
    state['q'] = state.Tatm * 0. + Q
    state['U'] = state.Tatm * 0. + U
    state['V'] = state.Tatm * 0. + V
    assert hasattr(state, 'Tatm')
    assert hasattr(state, 'q')
    assert hasattr(state, 'U')
    assert hasattr(state, 'V')
    conv = emanuel_convection.EmanuelConvection(state=state, timestep=DELT)
    conv.step_forward()
    #  Did we get all the correct output?
    assert conv.IFLAG == 1
    #  relative tolerance for these tests ...
    tol = 1E-5
    assert conv.CBMF == pytest.approx(CBMF, rel=tol)
    tend = conv.tendencies
    assert FT == pytest.approx(tend['Tatm'], rel=tol)
    assert FQ == pytest.approx(tend['q'], rel=tol)
    assert FU == pytest.approx(tend['U'], rel=tol)
    assert FV == pytest.approx(tend['V'], rel=tol)

@pytest.mark.compiled
@pytest.mark.fast
def test_multidim_tendencies():
    # Same test just repeated in two parallel columns
    num_lat = 2
    state = climlab.column_state(num_lev=num_lev, num_lat=num_lat)
    state['q'] = state.Tatm * 0. #+ Q
    state['U'] = state.Tatm * 0. #+ U
    state['V'] = state.Tatm * 0. #+ V
    for i in range(num_lat):
        state.Tatm[i,:] = T
        state['q'][i,:] += Q
        state['U'][i,:] += U
        state['V'][i,:] += V
    assert hasattr(state, 'Tatm')
    assert hasattr(state, 'q')
    assert hasattr(state, 'U')
    assert hasattr(state, 'V')
    conv = emanuel_convection.EmanuelConvection(state=state, timestep=DELT)
    conv.step_forward()
    #  Did we get all the correct output?
    assert np.all(conv.IFLAG == 1)
    #  relative tolerance for these tests ...
    tol = 1E-5
    assert np.all(conv.CBMF == pytest.approx(CBMF, rel=tol))
    tend = conv.tendencies
    assert np.tile(FT,(num_lat,1)) == pytest.approx(tend['Tatm'], rel=tol)
    assert np.tile(FQ,(num_lat,1)) == pytest.approx(tend['q'], rel=tol)
    assert np.tile(FU,(num_lat,1)) == pytest.approx(tend['U'], rel=tol)
    assert np.tile(FV,(num_lat,1)) == pytest.approx(tend['V'], rel=tol)

@pytest.mark.compiled
@pytest.mark.fast
def test_rcm_emanuel():
    num_lev = 30
    water_depth = 5.
    # Temperatures in a single column
    state = climlab.column_state(num_lev=num_lev, water_depth=water_depth)
    #  Initialize a nearly dry column (small background stratospheric humidity)
    state['q'] = np.ones_like(state.Tatm) * 5.E-6
    #  ASYNCHRONOUS COUPLING -- the radiation uses a much longer timestep
    short_timestep = climlab.constants.seconds_per_hour
    #  The top-level model
    model = climlab.TimeDependentProcess(name='Radiative-Convective Model',
                        state=state,
                        timestep=short_timestep)
    #  Radiation coupled to water vapor
    rad = climlab.radiation.RRTMG(name='Radiation',
                        state=state,
                        specific_humidity=state.q,
                        albedo=0.3,
                        timestep=24*short_timestep)
    #  Convection scheme -- water vapor is a state variable
    conv = climlab.convection.EmanuelConvection(name='Convection',
                                  state=state,
                                  timestep=short_timestep)
    #  Surface heat flux processes
    shf = climlab.surface.SensibleHeatFlux(name='SHF',
                                  state=state, Cd=0.5E-3,
                                  timestep=climlab.constants.seconds_per_hour)
    lhf = climlab.surface.LatentHeatFlux(name='LHF',
                                  state=state, Cd=0.5E-3,
                                  timestep=short_timestep)
    #  Couple all the submodels together
    for proc in [rad, conv, shf, lhf]:
        model.add_subprocess(proc.name, proc)
    model.step_forward()
    to_xarray(model)



"""
A subclass of MutableAttr that has defaultdict support.
"""
from collections.abc import Mapping

import six

from .mixins import MutableAttr


__all__ = ['AttrDefault']


class AttrDefault(MutableAttr):
    """
    An implementation of MutableAttr with defaultdict support
    """
    def __init__(self, default_factory=None, items=None, sequence_type=tuple,
                 pass_key=False):
        if items is None:
            items = {}
        elif not isinstance(items, Mapping):
            items = dict(items)

        self._setattr('_default_factory', default_factory)
        self._setattr('_mapping', items)
        self._setattr('_sequence_type', sequence_type)
        self._setattr('_pass_key', pass_key)
        self._setattr('_allow_invalid_attributes', False)

    def _configuration(self):
        """
        The configuration for a AttrDefault instance
        """
        return self._sequence_type, self._default_factory, self._pass_key

    def __getitem__(self, key):
        """
        Access a value associated with a key.

        Note: values returned will not be wrapped, even if recursive
        is True.
        """
        if key in self._mapping:
            return self._mapping[key]
        elif self._default_factory is not None:
            return self.__missing__(key)

        raise KeyError(key)

    def __setitem__(self, key, value):
        """
        Add a key-value pair to the instance.
        """
        self._mapping[key] = value

    def __delitem__(self, key):
        """
        Delete a key-value pair
        """
        del self._mapping[key]

    def __len__(self):
        """
        Check the length of the mapping.
        """
        return len(self._mapping)

    def __iter__(self):
        """
        Iterated through the keys.
        """
        return iter(self._mapping)

    def __missing__(self, key):
        """
        Add a missing element.
        """
        if self._pass_key:
            self[key] = value = self._default_factory(key)
        else:
            self[key] = value = self._default_factory()

        return value

    def __repr__(self):
        """
        Return a string representation of the object.
        """
        return six.u(
            "AttrDefault({default_factory}, {pass_key}, {mapping})"
        ).format(
            default_factory=repr(self._default_factory),
            pass_key=repr(self._pass_key),
            mapping=repr(self._mapping),
        )

    def __getstate__(self):
        """
        Serialize the object.
        """
        return (
            self._default_factory,
            self._mapping,
            self._sequence_type,
            self._pass_key,
            self._allow_invalid_attributes,
        )

    def __setstate__(self, state):
        """
        Deserialize the object.
        """
        (default_factory, mapping, sequence_type, pass_key,
         allow_invalid_attributes) = state

        self._setattr('_default_factory', default_factory)
        self._setattr('_mapping', mapping)
        self._setattr('_sequence_type', sequence_type)
        self._setattr('_pass_key', pass_key)
        self._setattr('_allow_invalid_attributes', allow_invalid_attributes)

    @classmethod
    def _constructor(cls, mapping, configuration):
        """
        A standardized constructor.
        """
        sequence_type, default_factory, pass_key = configuration
        return cls(default_factory, mapping, sequence_type=sequence_type,
                   pass_key=pass_key)



r"""Solver for the 1D meridional moist static energy diffusion equation on the sphere:

.. math::

    C\frac{\partial}{\partial t} T(\phi,t) = \frac{1}{\cos\phi} \frac{\partial}{\partial \phi} \left[ \cos\phi ~ D ~(1+f(T))~ \frac{\partial T}{\partial \phi} \right]

where :math:`f(T)` is a temperature-dependent moisture amplification factor given by

.. math::

    f(T) = \frac{L^2 r q^*(T)}{c_p R_v T^2}

which expresses the effect of latent heat on the near-surface moist static energy,
where :math:`q^*(T)` is the saturation specific humidity at temperature :math:`T`
and :math:`r` is a relative humidity.

This class operates identically to ``MeridionalHeatDiffusion``
but calculates :math:`f`
automatically at each timestep and applies it to the diffusivity.

The magnitude of the moisture amplification is controlled by the input parameter
`relative_humidity` (i.e. :math:`r` in the equation above).

It can be used to implement a modified Energy Balance Model accounting for the
effects of moisture on the heat transport efficiency.


Derivation of the moist diffusion equation
------------------------------------------

Assume that heat transport is down the gradient of **moist static energy**
:math:`m = c_p T + L q + g Z`

For an EBM we want to parameterize everything in terms of a surface temperature :math:`T_s`.
So we write :math:`m_s = c_p T_s + L r q^*(T_s)`,
where :math:`m_s` is the moist static energy of near-surface air parcels,
:math:`r` is a near-surface relative humidity,
and :math:`q^*` is the **saturation specific humidity** at a reference surface pressure.

Now express this quantity in temperature units by defining a *moist temperature*

.. math::

    T_m = \frac{m_s}{c_p} = T_s + \frac{L r}{c_p} q^*(T_s)

:math:`T_m` is the temperature a dry air parcel would have
that has the same total enthalpy as a moist air parcel at temperature :math:`T_s`

The down-gradient heat transport parameterization can then be written

.. math::
    \mathcal{H} = -2 \pi a^2 D_m \frac{\partial T_m}{\partial \phi}

where :math:`D_m` is the thermal diffusion coefficient for this moist model, in units of W/m2/K.

The equation we are trying to solve is thus

.. math::

    C \frac{\partial T_s}{\partial t} = \frac{1}{\cos\phi} \frac{\partial}{\partial \phi} \left( \cos\phi D_m \frac{\partial T_m}{\partial \phi} \right)

which we can write in terms of :math:`T_s` only by substituting in for :math:`T_m`:

.. math::

    C \frac{\partial T_s}{\partial t} = \frac{1}{\cos\phi} \frac{\partial}{\partial \phi} \left( \cos\phi D_m \left(\frac{\partial T_s}{\partial \phi} + \frac{\partial}{\partial \phi} \left(\frac{L r}{c_p} q^*(T_s)\right)\right)\right)

If we make the simplifying assumption that the **relative humidity :math:`r` is constant**
(not a function of latitude), then

.. math::

    C \frac{\partial T_s}{\partial t} = \frac{1}{\cos\phi} \frac{\partial}{\partial \phi} \left( \cos\phi D_m \left(\frac{\partial T_s}{\partial \phi} + \frac{L r}{c_p} \frac{\partial q^*}{\partial \phi} \right)\right)

To a good approximation (see Hartmann's book and others),
the Clausius-Clapeyron relation for saturation specific humidity gives

.. math::

    \frac{\partial q^*}{dT} = \frac{L}{R_v T^2} q^*(T)

Then using a chain rule we have

.. math::

    \frac{\partial q^*}{\partial \phi} = \frac{\partial q^*}{\partial T_s} \frac{\partial T_s}{\partial \phi}  = \frac{L q^*(T_s)}{R_v T_s^2}  \frac{\partial T_s}{\partial \phi}

Plugging this into our model equation we get

.. math::

    C \frac{\partial T_s}{\partial t} = \frac{1}{\cos\phi} \frac{\partial}{\partial \phi} \left( \cos\phi D_m \frac{\partial T_s}{\partial \phi} \left(1 + \frac{L^2 r q^*(T_s)}{c_p R_v T_s^2} \right)\right)

This is now in a form that is compatible with our diffusion solver.

Just let

.. math::

    D = D_m \left( 1 + f(T_s) \right)

where

.. math::

    f(T_s) = \frac{L^2 r q^*(T_s)}{c_p R_v T_s^2}

or, equivalently,

.. math::

    f(T_s) = \frac{L r }{c_p} \frac{\partial q^*}{dT}\bigg|_{T_s}

Given a temperature distribution :math:`T_s(\phi)` at any given time,
we can calculate the diffusion coefficient :math:`D(\phi)` from this formula.

This calculation is implemented in the ``MeridionalMoistDiffusion`` class.
"""
from __future__ import division
import numpy as np
from .meridional_heat_diffusion import MeridionalHeatDiffusion
from climlab.utils.thermo import qsat
from climlab import constants as const


class MeridionalMoistDiffusion(MeridionalHeatDiffusion):
    def __init__(self, D=0.24, relative_humidity=0.8, **kwargs):
        self.relative_humidity = relative_humidity
        super(MeridionalMoistDiffusion, self).__init__(D=D, **kwargs)
        self._update_diffusivity()

    def _update_diffusivity(self):
        Tinterp = np.interp(self.lat_bounds, self.lat, np.squeeze(self.Ts))
        Tkelvin = Tinterp + const.tempCtoK
        f = moist_amplification_factor(Tkelvin, self.relative_humidity)
        heat_capacity = self.Ts.domain.heat_capacity
        self.K = self.D / heat_capacity * const.a**2 * (1+f)

    def _implicit_solver(self):
        self._update_diffusivity()
        #  and then do all the same stuff the parent class would do...
        return super(MeridionalMoistDiffusion, self)._implicit_solver()


def moist_amplification_factor(Tkelvin, relative_humidity=0.8):
    '''Compute the moisture amplification factor for the moist diffusivity
    given relative humidity and reference temperature profile.'''
    deltaT = 0.01
    #  slope of saturation specific humidity at 1000 hPa
    dqsdTs = (qsat(Tkelvin+deltaT/2, 1000.) - qsat(Tkelvin-deltaT/2, 1000.)) / deltaT
    return const.Lhvap / const.cp * relative_humidity * dqsdTs



from __future__ import division
import numpy as np
from scipy.interpolate import interp1d
from climlab.utils.thermo import mmr_to_vmr


def _prepare_general_arguments(RRTMGobject):
    '''Prepare arguments needed for both RRTMG_SW and RRTMG_LW with correct dimensions.'''
    tlay = _climlab_to_rrtm(RRTMGobject.Tatm)
    tlev = _climlab_to_rrtm(interface_temperature(**RRTMGobject.state))
    play = _climlab_to_rrtm(RRTMGobject.lev * np.ones_like(tlay))
    plev = _climlab_to_rrtm(RRTMGobject.lev_bounds * np.ones_like(tlev))
    ncol, nlay = tlay.shape
    tsfc = _climlab_to_rrtm_sfc(RRTMGobject.Ts, RRTMGobject.Ts)
    # GASES -- put them in proper dimensions and units
    vapor_mixing_ratio = mmr_to_vmr(RRTMGobject.specific_humidity, gas='H2O')
    h2ovmr   = _climlab_to_rrtm(vapor_mixing_ratio * np.ones_like(RRTMGobject.Tatm))
    o3vmr    = _climlab_to_rrtm(RRTMGobject.absorber_vmr['O3'] * np.ones_like(RRTMGobject.Tatm))
    co2vmr   = _climlab_to_rrtm(RRTMGobject.absorber_vmr['CO2'] * np.ones_like(RRTMGobject.Tatm))
    ch4vmr   = _climlab_to_rrtm(RRTMGobject.absorber_vmr['CH4'] * np.ones_like(RRTMGobject.Tatm))
    n2ovmr   = _climlab_to_rrtm(RRTMGobject.absorber_vmr['N2O'] * np.ones_like(RRTMGobject.Tatm))
    o2vmr    = _climlab_to_rrtm(RRTMGobject.absorber_vmr['O2'] * np.ones_like(RRTMGobject.Tatm))
    cfc11vmr = _climlab_to_rrtm(RRTMGobject.absorber_vmr['CFC11'] * np.ones_like(RRTMGobject.Tatm))
    cfc12vmr = _climlab_to_rrtm(RRTMGobject.absorber_vmr['CFC12'] * np.ones_like(RRTMGobject.Tatm))
    cfc22vmr = _climlab_to_rrtm(RRTMGobject.absorber_vmr['CFC22'] * np.ones_like(RRTMGobject.Tatm))
    ccl4vmr  = _climlab_to_rrtm(RRTMGobject.absorber_vmr['CCL4'] * np.ones_like(RRTMGobject.Tatm))
    #  Cloud parameters
    cldfrac = _climlab_to_rrtm(RRTMGobject.cldfrac * np.ones_like(RRTMGobject.Tatm))
    ciwp = _climlab_to_rrtm(RRTMGobject.ciwp * np.ones_like(RRTMGobject.Tatm))
    clwp = _climlab_to_rrtm(RRTMGobject.clwp * np.ones_like(RRTMGobject.Tatm))
    relq = _climlab_to_rrtm(RRTMGobject.r_liq * np.ones_like(RRTMGobject.Tatm))
    reic = _climlab_to_rrtm(RRTMGobject.r_ice * np.ones_like(RRTMGobject.Tatm))

    return (ncol, nlay, play, plev, tlay, tlev, tsfc,
            h2ovmr, o3vmr, co2vmr, ch4vmr, n2ovmr, o2vmr, cfc11vmr,
            cfc12vmr, cfc12vmr, cfc22vmr, ccl4vmr,
            cldfrac, ciwp, clwp, relq, reic)



def interface_temperature(Ts, Tatm, **kwargs):
    '''Compute temperature at model layer interfaces.'''
    #  Actually it's not clear to me how the RRTM code uses these values
    lev = Tatm.domain.axes['lev'].points
    lev_bounds = Tatm.domain.axes['lev'].bounds
    #  Interpolate to layer interfaces
    f = interp1d(lev, Tatm, axis=-1)  # interpolation function
    Tinterp = f(lev_bounds[1:-1])
    #  add TOA value, Assume surface temperature at bottom boundary
    Ttoa = Tatm[...,0]
    Tinterp = np.concatenate((Ttoa[..., np.newaxis], Tinterp, Ts), axis=-1)
    return Tinterp

def _climlab_to_rrtm(field):
    '''Prepare field with proper dimension order.
    RRTM code expects arrays with (ncol, nlay)
    and with pressure decreasing from surface at element 0

    climlab grid dimensions are any of:
        - (num_lev,) --> (1, num_lev)
        - (num_lat, num_lev)  --> (num_lat, num_lev)
        - (num_lat, num_lon, num_lev)  -->  (num_lat*num_lon, num_lev)

        But lat-lon grids not yet supported here!

    Case single column
    '''
    # Make this work just with 1D (KM,) arrays
    #  (KM,)  -->  (1, nlay)
    try:
        #  Flip along the last axis to reverse the pressure order
        field = field[..., ::-1]
    except:
        if np.isscalar(field):
            return field
        else:
            raise ValueError('field must be array_like or scalar.')
    shape = field.shape
    if len(shape)==1:  #  (num_lev)
        #  Need to append an extra dimension for singleton horizontal ncol
        return field[np.newaxis, ...]
    elif len(shape)==2:  # (num_lat, num_lev)
        return field
    elif len(shape) > 2:
        raise ValueError('lat-lon grids not yet supported here.')
    #elif len(shape)==3:  # (num_lat, num_lon, num_lev)
        #  Need to reshape this array

def _rrtm_to_climlab(field):
    try:
        #  Flip along the last axis to reverse the pressure order
        field = field[..., ::-1]
    except:
        if np.isscalar(field):
            return field
        else:
            raise ValueError('field must be array_like or scalar.')
    return np.squeeze(field)

def _climlab_to_rrtm_sfc(field, Ts):
    '''Return an array of size np.squeeze(Ts) to remove the singleton depth dimension'''
    fieldsqueeze = np.squeeze(field)
    Tsqueeze = np.squeeze(Ts)
    return fieldsqueeze * np.ones_like(Tsqueeze)



from __future__ import division
from builtins import range
import numpy as np
from climlab.utils.thermo import blackbody_emission
from climlab.radiation.transmissivity import Transmissivity
from climlab.process import EnergyBudget


class GreyGas(EnergyBudget):
    '''Base class for all band radiation models,
    including grey and semi-grey model.

    Input argument absorptivity is band absorptivity
    (should same size as grid).

    By default emissivity = absorptivity.
    Subclasses can override this is necessary (e.g. for shortwave model).

    The following boundary and input values need to be specified by
    user or parent process:
    - albedo_sfc (default is zero)
    - flux_from_space
    - absorptivity
    - reflectivity (default is zero)
    These are accessible (and settable) as process attributes
    Also stored in process.input dictionary

    The following values are computed are stored in the .diagnostics dictionary:
    - flux_from_sfc
    - flux_to_sfc
    - flux_to_space
    - absorbed
    - absorbed_total
    - emission
    - emission_sfc
    - flux_reflected_up
    (all in W/m2)
    '''
    def __init__(self, absorptivity=None, reflectivity=None, emissivity_sfc=1.,
                 albedo_sfc=0., **kwargs):
        super(GreyGas, self).__init__(**kwargs)
        self.add_input('flux_from_space', 0. * self.Ts)
        #  initialize all diagnostics to zero
        self.add_diagnostic('flux_from_sfc', 0. * self.Ts)
        self.add_diagnostic('flux_to_sfc', 0. * self.Ts)
        self.add_diagnostic('flux_to_space', 0. * self.Ts)
        self.add_diagnostic('absorbed', 0. * self.Tatm)
        self.add_diagnostic('absorbed_total', 0. * self.Ts)

        newinput = ['reflectivity',
                    'absorptivity',
                    'emissivity_sfc',
                    'albedo_sfc',]
        self.declare_input(newinput)
        if reflectivity is None:
            reflectivity = np.zeros_like(self.Tatm)
        if absorptivity is None:
            absorptivity = np.zeros_like(self.Tatm)
        self.absorptivity = absorptivity
        self.reflectivity = reflectivity
        self.emissivity_sfc = emissivity_sfc * np.ones_like(self.Ts)
        self.albedo_sfc = albedo_sfc * np.ones_like(self.Ts)
        #  Initialize diagnostics
        self.add_diagnostic('emission', 0. * self.Tatm)
        self.add_diagnostic('emission_sfc', 0. * self.Ts)
        self.add_diagnostic('flux_reflected_up')

    @property
    def absorptivity(self):
        return self.trans.absorptivity
    @absorptivity.setter
    def absorptivity(self, value):
        #  value should be a Field,
         #  or numpy array of same size as self.Tatm
        try:
            axis = value.domain.axis_index['lev']
        except:
            axis = self.Tatm.domain.axis_index['lev']
            # if a single scalar is given, broadcast that to all levels
            if len(np.shape(np.array(value))) == 0:
                value = np.ones_like(self.Tatm) * value
            elif value.shape != self.Tatm.shape:
                raise ValueError('absorptivity must be a Field, a scalar, or match atm grid dimensions')
        try:
            self.trans = Transmissivity(absorptivity=value,
                                    reflectivity=self.reflectivity,
                                    axis=axis)
        except:
            self.trans = Transmissivity(absorptivity=value, axis=axis)
        self.input['absorptivity'] = value
    @property
    def emissivity(self):
        # This ensures that emissivity = absorptivity at all times
    #  needs to be overridden for shortwave classes
        return self.absorptivity
    @property
    def transmissivity(self):
        return self.trans.transmissivity
    @transmissivity.setter
    def transmissivity(self, value):
        self.absorptivity = 1 - value
    @property
    def reflectivity(self):
        return self.trans.reflectivity
    @reflectivity.setter
    def reflectivity(self, value):
        #  value should be a Field,
         #  or numpy array of same size as self.Tatm
        try:
            axis = value.domain.axis_index['lev']
        except:
            axis = self.Tatm.domain.axis_index['lev']
            # if a single scalar is given, broadcast that to all levels
            if len(np.shape(np.array(value))) == 0:
                value = np.ones_like(self.Tatm) * value
            elif value.shape != self.Tatm.shape:
                raise ValueError('reflectivity must be a Field, a scalar, or match atm grid dimensions')
        self.trans = Transmissivity(absorptivity=self.absorptivity,
                                    reflectivity=value,
                                    axis=axis)
        #self.input['reflectivity'] = value

    def _compute_emission_sfc(self):
        return self.emissivity_sfc * blackbody_emission(self.Ts)

    def _compute_emission(self):
        return self.emissivity * blackbody_emission(self.Tatm)

    def _compute_fluxes(self):
        ''' All fluxes are band by band'''
        self.emission = self._compute_emission()
        self.emission_sfc = self._compute_emission_sfc()
        fromspace = self._from_space()
        self.flux_down = self.trans.flux_down(fromspace, self.emission)
        self.flux_reflected_up = self.trans.flux_reflected_up(self.flux_down, self.albedo_sfc)
        # this ensure same dimensions as other fields
        self.flux_to_sfc = self.flux_down[..., -1, np.newaxis]
        self.flux_from_sfc = (self.emission_sfc +
                              self.flux_reflected_up[..., -1, np.newaxis])
        self.flux_up = self.trans.flux_up(self.flux_from_sfc,
                            self.emission + self.flux_reflected_up[...,0:-1])
        self.flux_net = self.flux_up - self.flux_down
        # absorbed radiation (flux convergence) in W / m**2 (per band)
        self.absorbed = np.diff(self.flux_net, axis=-1)
        self.absorbed_total = np.sum(self.absorbed, axis=-1)
        self.flux_to_space = self._compute_flux_top()

    def _compute_flux_top(self):
        bandflux = self.flux_up[..., 0, np.newaxis]
        return self._join_channels(bandflux)

    def _flux_convergence_atm(self):
        return self.absorbed

    def _flux_convergence_sfc(self):
        return ( self.flux_to_sfc - self.flux_from_sfc )

    def _compute_heating_rates(self):
        '''Compute energy flux convergences to get heating rates in :math:`W/m^2`.'''
        self._compute_radiative_heating()

    def _compute_radiative_heating(self):
        self._compute_fluxes()
        self.heating_rate['Tatm'] = self._join_channels(self.absorbed)
        self.heating_rate['Ts'] = self._join_channels( self.flux_to_sfc -
                                                       self.flux_from_sfc )

    def _from_space(self):
        try:
            fromspace = self.flux_from_space
        except:
            fromspace = np.zeros_like(self.Ts)
        return self._split_channels(fromspace)

    def _split_channels(self, flux):
        '''Single channel for Grey Gas model.'''
        return flux

    def _join_channels(self, flux):
        '''Single channel for Grey Gas model.'''
        return flux


#####  These routines may need fixing with flipped vertical grid...
    def flux_components_top(self):
        '''Compute the contributions to the outgoing flux to space due to
        emissions from each level and the surface.'''
        N = self.lev.size
        flux_up_bottom = self.flux_from_sfc
        emission = np.zeros_like(self.emission)
        this_flux_up = (np.ones_like(self.Ts) *
                        self.trans.flux_up(flux_up_bottom, emission))
        sfcComponent = this_flux_up[..., -1]
        atmComponents = np.zeros_like(self.Tatm)
        flux_up_bottom = np.zeros_like(self.Ts)
        # I'm sure there's a way to write this as a vectorized operation
        #  but the speed doesn't really matter if it's just for diagnostic
        #  and we are not calling it every timestep
        for n in range(N):
            emission = np.zeros_like(self.emission)
            emission[..., n] = self.emission[..., n]
            this_flux_up = self.trans.flux_up(flux_up_bottom, emission)
            atmComponents[..., n] = this_flux_up[..., -1]
        return sfcComponent, atmComponents

    def flux_components_bottom(self):
        '''Compute the contributions to the downwelling flux to surface due to
        emissions from each level.'''
        N = self.lev.size
        atmComponents = np.zeros_like(self.Tatm)
        flux_down_top = np.zeros_like(self.Ts)
        #  same comment as above... would be nice to vectorize
        for n in range(N):
            emission = np.zeros_like(self.emission)
            emission[..., n] = self.emission[..., n]
            this_flux_down = self.trans.flux_down(flux_down_top, emission)
            atmComponents[..., n] = this_flux_down[..., 0]
        return atmComponents


class GreyGasSW(GreyGas):
    '''Emissivity is always set to zero for shortwave classes.'''
    def __init__(self, albedo_sfc=0.33, emissivity_sfc=0., **kwargs):
        super(GreyGasSW, self).__init__(albedo_sfc=albedo_sfc,
                                        emissivity_sfc=emissivity_sfc,
                                        **kwargs)
    @property
    def emissivity(self):
        # This ensures that emissivity is always zero for shortwave classes
        return np.zeros_like(self.absorptivity)



from __future__ import division, print_function
import numpy as np
import os, pooch
import pandas as pd
import xarray as xr


base_url = 'http://vo.imcce.fr/insola/earth/online/earth/La2004/'
filenames = {'past': 'INSOLN.LA2004.BTL.ASC',
             'future': 'INSOLP.LA2004.BTL.ASC'}
hashes = {'past': "3f13b9f8e69085baf40bc67a2669e6f6af4148fef9218ed2158772aa91e35f8c",
          'future': "8e5ac423374802a4ce2a0958672271ed408054cc70214056c0c221c3d5b14750"}

def _get_Laskar_data(verbose=True):
    longorbithandle = {}
    longorbit = {}
    sources = {}
    pandas_kwargs = {'delim_whitespace':True,
                     'header':None,
                     'index_col':0,
                     'names':['kyear','ecc','obliquity','long_peri'],}
    for time in filenames:
        remote_path = base_url + filenames[time]
        if time == 'future':
            pandas_kwargs['skiprows'] = 1 # first row is kyear=0, redundant
        path = remote_path
        longorbithandle[time] = pooch.retrieve(url=path, known_hash=hashes[time])
        longorbit[time] = pd.read_csv(longorbithandle[time], **pandas_kwargs)
        sources[time] = path
    xlongorbit = {}
    for time in ['past', 'future']:
        # Cannot convert to float until we replace the D notation with E for floating point numbers
        longorbit[time].replace(to_replace='D', value='E', regex=True, inplace=True)
        xlongorbit[time] = xr.Dataset()
        xlongorbit[time]['ecc'] = xr.DataArray(pd.to_numeric(longorbit[time]['ecc']))
        for field in ['obliquity', 'long_peri']:
            xlongorbit[time][field] = xr.DataArray(np.rad2deg(pd.to_numeric(longorbit[time][field])))
    longorbit = xr.concat([xlongorbit['past'], xlongorbit['future']], dim='kyear')
    # add 180 degrees to long_peri (see lambda definition, Berger 1978 Appendix)
    longorbit['long_peri'] += 180.
    longorbit['precession'] = longorbit.ecc*np.sin(np.deg2rad(longorbit.long_peri))
    longorbit.attrs['Description'] = 'The Laskar et al. (2004) orbital data table'
    longorbit.attrs['Citation'] = 'https://doi.org/10.1051/0004-6361:20041335'
    longorbit.attrs['Source'] = [sources[time] for time in sources]
    longorbit.attrs['Note'] = 'Longitude of perihelion is defined to be 0 degrees at Northern Vernal Equinox. This differs by 180 degrees from the source files.'
    return longorbit

OrbitalTable = _get_Laskar_data()



from __future__ import division, print_function
import numpy as np
import os, pooch
import pandas as pd
import xarray as xr
from climlab.utils import _datapath_http


#  Possible sources for the Berger and Loutre 1991 data table
#NCDCpath = "https://www1.ncdc.noaa.gov/pub/data/paleo/climate_forcing/orbital_variations/insolation/orbit91"
path = _datapath_http + 'orbital/orbit91'

def _get_Berger_data(verbose=True):
    '''Read in the Berger and Loutre orbital table as a pandas dataframe, convert to xarray
    '''
    orbit91handle = pooch.retrieve(path,
        known_hash="3afc20dda7b385bdd366bc4c9cf60be02d8defdb4c0f317430ca8386d62f81a3")
    orbit91_pd = pd.read_csv(orbit91handle, delim_whitespace=True, skiprows=1, verbose=True)
    #  As xarray structure with the dimension named 'kyear'
    orbit = xr.Dataset(orbit91_pd).rename({'dim_0': 'kyear'})
    #  Now change names
    orbit = orbit.rename({'ECC': 'ecc', 'OMEGA': 'long_peri',
                          'OBL': 'obliquity', 'PREC': 'precession'})
    # add 180 degrees to long_peri (see lambda definition, Berger 1978 Appendix)
    orbit['long_peri'] += 180.
    # apply np.unwrap to remove discontinuities in the longitude of perihelion
    orbit['long_peri'] = np.rad2deg(xr.apply_ufunc(np.unwrap, np.deg2rad(orbit.long_peri)))
    orbit['precession'] *= -1.

    orbit.attrs['Description'] = 'The Berger and Loutre (1991) orbital data table'
    orbit.attrs['Citation'] = 'https://doi.org/10.1016/0277-3791(91)90033-Q'
    orbit.attrs['Source'] = path
    orbit.attrs['Note'] = 'Longitude of perihelion is defined to be 0 degrees at Northern Vernal Equinox. This differs by 180 degrees from orbit91 source file.'
    return orbit



'''The base classes for all climlab processes.'''
from __future__ import absolute_import
from .process import Process, process_like, get_axes
from .time_dependent_process import TimeDependentProcess, couple
from .implicit import ImplicitProcess
from .diagnostic import DiagnosticProcess
from .energy_budget import EnergyBudget
from .external_forcing import ExternalForcing



#!/usr/bin/env python
from __future__ import division, print_function


def configuration(parent_package='',top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('rrtm', parent_package, top_path)
    #config.make_config_py() # installs __config__.py
    return config

if __name__ == '__main__':
    print('This is the wrong setup.py file to run')



'''``_Radiation``, ``_Radiation_SW`` and ``_Radiation_LW``
are the base classes for radiative transfer modules

Currently this includes :class:`~climlab.radiation.cam3.CAM3`,
:class:`~climlab.radiation.rrtm.RRTMG`,
:class:`~climlab.radiation.rrtm.RRTMG_LW`, and
:class:`~climlab.radiation.rrtm.RRTMG_SW`

Basic characteristics:

State:

    - ``Ts`` (surface radiative temperature)
    - ``Tatm`` (air temperature)

Input arguments (both LW and SW):

    - ``specific_humidity`` (kg/kg)
    - ``absorber_vmr = None`` (dictionary of volumetric mixing ratios. Default values supplied if ``None``)
    - ``cldfrac``      (layer cloud fraction
    - ``clwp``         (in-cloud liquid water path (g/m2))
    - ``ciwp = 0.``,     # in-cloud ice water path (g/m2)
    - ``r_liq = 0.``,    # Cloud water drop effective radius (microns)
    - ``r_ice = 0.``,    # Cloud ice particle effective size (microns)
    - ``ozone_file = 'apeozone_cam3_5_54.nc'``  (file with ozone distribution - ignored if ``absorber_vmr`` is given)

If ``absorber_vmr = None`` then ozone will be interpolated to the model grid
from a climatology file, or set to zero if ``ozone_file = None``.

Additional input arguments for SW:

     - ``albedo = None``  (optional, single parameter to set all 4 albedo values)
     - ``aldif = 0.3``,   (near-infrared albedo, diffuse)
     - ``aldir = 0.3``,   (near-infrared albedo, direct)
     - ``asdif = 0.3``,   (shortwave albedo, diffuse)
     - ``asdir = 0.3``,   (shortwave albedo, direct)
     - ``S0 = const.S0``,  (solar constant, W/m2)
     - ``insolation = const.S0/4.``,  (time-mean insolaltion, W/m2)
     - ``coszen = None``,    # cosine of the solar zenith angle
     - ``eccentricity_factor = 1.``,  # instantaneous irradiance = S0 * eccentricity_factor

Additional input arguments for LW:
    - ``emissivity = 1.``,  # surface emissivity
    - ``return_spectral_olr = False``,  # Whether or not to return spectrally-decomposed Outgoing Longwave Radiation (implemented for RRTMG only)

Shortave processes compute these diagnostics (minimum):

    - ``ASR`` (W/m2, net Absorbed Shortwave Radiation at TOA, **positive down**)
    - ``ASRclr`` (clear-sky component)
    - ``ASRcld`` (cloud component, all-sky minus clear-sky)
    - ``SW_flux_up``   (W/m2, defined at pressure level interfaces)
    - ``SW_flux_down`` (W/m2, defined at pressure level interfaces)
    - ``SW_flux_net``  (W/m2 **downward** net flux at pressure level interfaces)
    - ``SW_flux_up_clr``  (clear-sky flux)
    - ``SW_flux_down_clr`` (clear-sky flux)
    - ``SW_flux_net_clr``  (clear-sky flux)
    - ``TdotSW``   (K/day, radiative heating rate)
    - ``TdotSW_clr`` (clear-sky heating rate)

Longwave processes compute these diagnostics (minimum):

    - ``OLR`` (W/m2, net Outgoing Longwave radiation at TOA, **positive up**)
    - ``OLRclr`` (clear-sky component)
    - ``OLRcld`` (cloud component, all-sky minus clear-sky)
    - ``LW_flux_up``   (W/m2, defined at pressure level interfaces)
    - ``LW_flux_down`` (W/m2, defined at pressure level interfaces)
    - ``LW_flux_net``  (W/m2 **upward** net flux at pressure level interfaces)
    - ``LW_flux_up_clr``  (clear-sky flux)
    - ``LW_flux_down_clr`` (clear-sky flux)
    - ``LW_flux_net_clr``  (clear-sky flux)
    - ``TdotLW``   (K/day, radiative heating rate)
    - ``TdotLW_clr`` (clear-sky heating rate)

If ``return_spectral_olr = True`` (RRTMG only), an additional diagnostic is produced:

    - ``OLR_spectral`` (W/m2, Outgoing Longwave Radiation at TOA in spectral bands, **positive up**)
'''

from __future__ import division
from __future__ import print_function
import numpy as np
from climlab.process import EnergyBudget
from climlab.radiation import ManabeWaterVapor
from climlab import constants as const
from climlab.domain.field import Field
import os, warnings, pooch
import xarray as xr
from climlab.utils import _datapath_http


def default_specific_humidity(Tatm):
    '''Initialize a specific humidity distribution
    based on a prescribed relative humidity profile.

    Input is air temperature array

    Output is specific humidity on same grid
    '''
    h2o = ManabeWaterVapor(state={'Tatm': Tatm})
    return h2o.q

def default_absorbers(Tatm,
                ozone_file = 'apeozone_cam3_5_54.nc',
                verbose = True,):
    '''Initialize a dictionary of well-mixed radiatively active gases
    All values are volumetric mixing ratios.

    Ozone is set to a climatology.

    All other gases are assumed well-mixed:

        - CO2
        - CH4
        - N2O
        - O2
        - CFC11
        - CFC12
        - CFC22
        - CCL4

    Specific values are based on the AquaPlanet Experiment protocols,
    except for O2 which is set the realistic value 0.21
    (affects the RRTMG scheme).
    '''
    absorber_vmr = {}
    absorber_vmr['CO2']   = 348. / 1E6
    absorber_vmr['CH4']   = 1650. / 1E9
    absorber_vmr['N2O']   = 306. / 1E9
    absorber_vmr['O2']    = 0.21
    absorber_vmr['CFC11'] = 0.
    absorber_vmr['CFC12'] = 0.
    absorber_vmr['CFC22'] = 0.
    absorber_vmr['CCL4']  = 0.

    # Ozone: start with all zeros, interpolate to data if we can
    xTatm = Tatm.to_xarray()
    O3 = 0. * xTatm
    if ozone_file is not None:
        ozonepath_http = _datapath_http + 'ozone/' + ozone_file
        ozonefilehandle = pooch.retrieve(url=ozonepath_http,
            known_hash="bc659bfa129fafa4ed9368bb19278ae15724a5a66599affd317c143ba511ff84")
        ozonedata = xr.open_dataset(ozonefilehandle)
        ##  zonal and time average
        ozone_zon = ozonedata.OZONE.mean(dim=('time','lon')).transpose('lat','lev')
        if ('lat' in xTatm.dims):
            O3source = ozone_zon
        else:
            weight = np.cos(np.deg2rad(ozonedata.lat))
            ozone_global = (ozone_zon * weight).mean(dim='lat') / weight.mean(dim='lat')
            O3source = ozone_global
        try:
            O3 = O3source.interp_like(xTatm)
            # There will be NaNs for gridpoints outside the ozone file domain
            assert not np.any(np.isnan(O3))
        except:
            warnings.warn('Some grid points are beyond the bounds of the ozone file. Ozone values will be extrapolated.')
            try:
                # passing fill_value='extrapolate' to the underlying scipy interpolator
                # will result in extrapolation instead of NaNs
                O3 = O3source.interp_like(xTatm, kwargs={'fill_value':'extrapolate'})
                assert not np.any(np.isnan(O3))
            except:
                warnings.warn('Interpolation of ozone data failed. Setting O3 to zero instead.')
                O3 = 0. * xTatm
    absorber_vmr['O3'] = O3.values
    return absorber_vmr

def init_interface(field):
    '''Return a Field object defined at the vertical interfaces of the input Field object.'''
    interface_shape = np.array(field.shape); interface_shape[-1] += 1
    interfaces = np.tile(False,len(interface_shape)); interfaces[-1] = True
    interface_zero = Field(np.zeros(interface_shape), domain=field.domain, interfaces=interfaces)
    return interface_zero


class _Radiation(EnergyBudget):
    '''Abstact base class for SW and LW radiation processes.
    '''
    def __init__(self,
            specific_humidity = None,
            #  Absorbing gases, volume mixing ratios
            absorber_vmr = None,
            cldfrac = 0.,  # layer cloud fraction
            clwp = 0.,     # in-cloud liquid water path (g/m2)
            ciwp = 0.,     # in-cloud ice water path (g/m2)
            r_liq = 0.,    # Cloud water drop effective radius (microns)
            r_ice = 0.,    # Cloud ice particle effective size (microns)
            ozone_file = 'apeozone_cam3_5_54.nc',
            **kwargs):
        super(_Radiation, self).__init__(**kwargs)
        #  Define inputs
        if specific_humidity is None:
            specific_humidity = default_specific_humidity(self.Tatm)
        self.add_input('specific_humidity', specific_humidity)
        if absorber_vmr is None:
            absorber_vmr = default_absorbers(self.Tatm, ozone_file, self.verbose)
        self.add_input('absorber_vmr', absorber_vmr)
        self.add_input('cldfrac', cldfrac)
        self.add_input('clwp', clwp)
        self.add_input('ciwp', ciwp)
        self.add_input('r_liq', r_liq)
        self.add_input('r_ice', r_ice)


class _Radiation_SW(_Radiation):
    '''Parent class for SW radiation modules
    '''
    def __init__(self,
                 albedo = None,
                 aldif = 0.3,
                 aldir = 0.3,
                 asdif = 0.3,
                 asdir = 0.3,
                 S0    = const.S0,
                 insolation = const.S0/4.,
                 coszen = None,    # cosine of the solar zenith angle
                 eccentricity_factor = 1.,  # instantaneous irradiance = S0 * eccentricity_factor
                 **kwargs):
        super(_Radiation_SW, self).__init__(**kwargs)
        #  coszen is cosine of solar zenith angle
        #  If unspecified, infer it from the insolation
        #  (assuming a circular orbit and standard solar constant)
        if coszen is None:
            coszen = insolation / S0
        self.add_input('S0', S0)
        self.add_input('insolation', insolation)
        self.add_input('coszen', coszen)
        self.add_input('eccentricity_factor', eccentricity_factor)
        if albedo is not None:
            aldif = albedo
            aldir = albedo
            asdif = albedo
            asdir = albedo
        self.add_input('aldif', aldif)
        self.add_input('aldir', aldir)
        self.add_input('asdif', asdif)
        self.add_input('asdir', asdir)
        # initialize diagnostics
        self.add_diagnostic('ASR', 0. * self.Ts)
        self.add_diagnostic('ASRclr', 0. * self.Ts)
        self.add_diagnostic('ASRcld', 0. * self.Ts)
        self.add_diagnostic('TdotSW', 0. * self.Tatm)
        self.add_diagnostic('TdotSW_clr', 0.*self.Tatm)
        self.add_diagnostic('SW_sfc', 0.*self.Ts)
        self.add_diagnostic('SW_sfc_clr', 0.*self.Ts)
        #  Flux diagnostics at layer interfaces
        #   These need an extra vertical level
        interface_zero = init_interface(self.Tatm)
        self.add_diagnostic('SW_flux_up', 0. * interface_zero)
        self.add_diagnostic('SW_flux_down', 0. * interface_zero)
        self.add_diagnostic('SW_flux_net', 0. * interface_zero)
        self.add_diagnostic('SW_flux_up_clr', 0. * interface_zero)
        self.add_diagnostic('SW_flux_down_clr', 0. * interface_zero)
        self.add_diagnostic('SW_flux_net_clr', 0. * interface_zero)

    def _compute_SW_flux_diagnostics(self):
        #  positive down, consistent with ASR
        self.SW_flux_net = self.SW_flux_down - self.SW_flux_up
        self.SW_flux_net_clr = self.SW_flux_down_clr - self.SW_flux_up_clr
        #  TOA diagnostics
        self.ASR = np.array(self.SW_flux_net[..., 0, np.newaxis]) + 0.*self.Ts
        self.ASRclr = np.array(self.SW_flux_net_clr[..., 0, np.newaxis]) + 0.*self.Ts
        self.ASRcld = self.ASR - self.ASRclr
        #  Surface diagnostics
        self.SW_sfc = np.array(self.SW_flux_net[..., -1, np.newaxis]) + 0.*self.Ts
        self.SW_sfc_clr = np.array(self.SW_flux_net_clr[..., -1, np.newaxis]) + 0.*self.Ts


class _Radiation_LW(_Radiation):
    def __init__(self,
                 emissivity = 1.,  # surface emissivity
                 **kwargs):
        super(_Radiation_LW, self).__init__(**kwargs)
        self.add_input('emissivity', emissivity)
        # initialize diagnostics
        self.add_diagnostic('OLR', 0. * self.Ts)
        self.add_diagnostic('OLRclr', 0. * self.Ts)
        self.add_diagnostic('OLRcld', 0. * self.Ts)
        self.add_diagnostic('TdotLW', 0. * self.Tatm)
        self.add_diagnostic('TdotLW_clr', 0.*self.Tatm)
        self.add_diagnostic('LW_sfc', 0.*self.Ts)
        self.add_diagnostic('LW_sfc_clr', 0.*self.Ts)

        #  Flux diagnostics at layer interfaces
        #   These need an extra vertical level
        interface_zero = init_interface(self.Tatm)
        self.add_diagnostic('LW_flux_up', 0. * interface_zero)
        self.add_diagnostic('LW_flux_down', 0. * interface_zero)
        self.add_diagnostic('LW_flux_net', 0. * interface_zero)
        self.add_diagnostic('LW_flux_up_clr', 0. * interface_zero)
        self.add_diagnostic('LW_flux_down_clr', 0. * interface_zero)
        self.add_diagnostic('LW_flux_net_clr', 0. * interface_zero)

    def _compute_LW_flux_diagnostics(self):
        #  LW net flux defined positive UP, consistent with OLR
        self.LW_flux_net = self.LW_flux_up - self.LW_flux_down
        self.LW_flux_net_clr = self.LW_flux_up_clr - self.LW_flux_down_clr
        #  TOA diagnostics
        self.OLR = np.array(self.LW_flux_net[..., 0, np.newaxis]) + 0.*self.Ts
        self.OLRclr = np.array(self.LW_flux_net_clr[..., 0, np.newaxis]) + 0.*self.Ts
        self.OLRcld = self.OLR - self.OLRclr
        #  Surface diagnostics
        self.LW_sfc = np.array(self.LW_flux_net[..., -1, np.newaxis]) + 0.*self.Ts
        self.LW_sfc_clr = np.array(self.LW_flux_net_clr[..., -1, np.newaxis]) + 0.*self.Ts



from __future__ import division
import numpy as np
from climlab import constants as const
import sys


def convective_adjustment_direct(p, T, c, lapserate=6.5):
    """Convective Adjustment to a specified lapse rate.

    Input argument lapserate gives the lapse rate expressed in degrees K per km
    (positive means temperature increasing downward).

    Default lapse rate is 6.5 K / km.

    Returns the adjusted Column temperature.
    inputs:
    p is pressure in hPa
    T is temperature in K
    c is heat capacity in in J / m**2 / K

    Implements the conservative adjustment algorithm from Akmaev (1991) MWR
    """
    # make sure lapserate has same dimensionality as T
    lapserate = lapserate * np.ones_like(T)
    # largely follows notation and algorithm in Akmaev (1991) MWR
    alpha = const.Rd / const.g * lapserate / 1.E3 # same dimensions as lapserate
    L = p.size
    ###  now handles variable lapse rate in multiple dimensions
    # prepend const.ps = 1000 hPa as ref pressure to compute potential temperature
    pextended = np.insert(p,0,const.ps)
    #  For now, let's assume that the vertical axis is the last axis
    Pi = np.cumprod((p / pextended[:-1])**alpha, axis=-1)  # Akmaev's equation 14 recurrence formula
    beta = 1./Pi
    theta = T * beta
    q = Pi * c
    n_k = np.zeros(L, dtype=np.int8)
    theta_k = np.zeros_like(p)
    s_k = np.zeros_like(p)
    t_k = np.zeros_like(p)
    thetaadj = Akmaev_adjustment_multidim(theta, q, beta, n_k,
                                           theta_k, s_k, t_k)
    T = thetaadj * Pi
    return T


def Akmaev_adjustment_multidim(theta, q, beta, n_k, theta_k, s_k, t_k):
    num_lev = theta.shape[-1]  # number of vertical levels
    otherdims = theta.shape[:-1] # everything except last dimension, which we assume is vertical
    if otherdims != ():
        othersize = np.prod(otherdims)
        theta_reshape = theta.reshape((othersize, num_lev))
        q_reshape = q.reshape((othersize, num_lev))
        beta_reshape = beta.reshape((othersize, num_lev))
        for n in range(othersize):
            theta_reshape[n,:] = Akmaev_adjustment(theta_reshape[n,:],
               q_reshape[n,:], beta_reshape[n,:], n_k, theta_k, s_k, t_k)
        theta = theta_reshape.reshape(theta.shape)
    else:
        theta = Akmaev_adjustment(theta, q, beta, n_k, theta_k, s_k, t_k)
    return theta


def Akmaev_adjustment(theta, q, beta, n_k, theta_k, s_k, t_k):
    '''Single column only.'''
    L = q.size  # number of vertical levels
    # Akmaev step 1
    k = 1
    n_k[k-1] = 1
    theta_k[k-1] = theta[k-1]
    l = 2
    while True:
        # Akmaev step 2
        n = 1
        thistheta = theta[l-1]
        while True:
            # Akmaev step 3
            if theta_k[k-1] <= thistheta:
                # Akmaev step 6
                k += 1
                break  # to step 7
            else:
                if n <= 1:
                    s = q[l-1]
                    t = s*thistheta
                # Akmaev step 4
                if n_k[k-1] <= 1:
                # lower adjacent level is not an earlier-formed neutral layer
                    s_k[k-1] = q[l-n-1]
                    t_k[k-1] = s_k[k-1] * theta_k[k-1]
                # Akmaev step 5
                    #  join current and underlying layers
                n += n_k[k-1]
                s += s_k[k-1]
                t += t_k[k-1]
                s_k[k-1] = s
                t_k[k-1] = t
                thistheta = t/s
                if k==1:
                    #  joint neutral layer is the first one
                    break  # to step 7
                k -= 1
                # back to step 3
        # Akmaev step 7
        if l == L:  # the scan is over
            break  # to step 8
        l += 1
        n_k[k-1] = n
        theta_k[k-1] = thistheta
        # back to step 2

    # update the potential temperatures
    while True:
        while True:
            # Akmaev step 8
            if n==1: # current model level was not included in any neutral layer
                break  # to step 11
            while True:
                #  Akmaev step 9
                theta[l-1] = thistheta
                if n==1:
                    break
                #  Akmaev step 10
                l -= 1
                n -= 1
                #  back to step 9
        # Akmaev step 11
        if k==1:
            break
        k -= 1
        l -= 1
        n = n_k[k-1]
        thistheta = theta_k[k-1]
        # back to step 8
    return theta

#  Attempt to use numba to compile the Akmaev_adjustment function
#  which gives at least 10x speedup
#   If numba is not available or compilation fails, the code will be executed
#   in pure Python. Results should be identical
try:
    from numba import jit
    Akmaev_adjustment = jit(signature_or_function=Akmaev_adjustment)
except ImportError:
    pass



from __future__ import division
import numpy as np
import climlab
import pytest
from climlab.tests.xarray_test import to_xarray
from climlab.utils.legendre import P2

@pytest.fixture()
def EBM_seasonal():
    return climlab.EBM_seasonal(water_depth=10.)

@pytest.fixture()
def EBM_highobliquity():
    orb_highobl = {'ecc':0., 'obliquity':90., 'long_peri':0.}
    return climlab.EBM_seasonal(orb=orb_highobl, water_depth=10.)

@pytest.fixture()
def EBM_iceline():
    return climlab.EBM_annual( num_points = 180, a0=0.3, a2=0.078, ai=0.62)

# helper for a common test pattern
def _check_minmax(array, amin, amax):
    return (np.allclose(array.min(), amin) and
            np.allclose(array.max(), amax))

@pytest.mark.fast
def test_model_creation(EBM_seasonal):
    """Just make sure we can create a model."""
    assert len(EBM_seasonal.Ts)==90

#  This is not a great test.. but just sees if we reproduce the same
#  temperature as in the notebook after 1 year of integration
@pytest.mark.fast
def test_integrate_years(EBM_seasonal):
    """Check that we can integrate forward the model and get the expected
    surface temperature."""
    EBM_seasonal.step_forward()
    EBM_seasonal.integrate_years(1)
    assert _check_minmax(EBM_seasonal.Ts, -24.21414189, 31.16169215)
    # Test the xarray interface
    to_xarray(EBM_seasonal)

#  And do the same thing at high obliquity
@pytest.mark.fast
def test_high_obliquity(EBM_highobliquity):
    """Check that we can integrate forward the model and get the expected
    surface temperature."""
    EBM_highobliquity.step_forward()
    EBM_highobliquity.integrate_years(1)
    assert _check_minmax(EBM_highobliquity.Ts, -9.28717079958807, 27.37890262018116)

@pytest.mark.fast
def test_annual_iceline(EBM_iceline):
    '''Check that the annual mean EBM with interactive ice edge gives expected
    result (ice at 70degrees).'''
    EBM_iceline.integrate_years(5.)
    assert np.all(EBM_iceline.icelat == np.array([-70.,  70.]))

@pytest.mark.fast
def test_decreased_S0(EBM_iceline):
    '''Check that a decrease in solar constant to 1200 W/m2 will give a
    Snowball Earth result in the annual mean EBM.'''
    #EBM_iceline.subprocess['insolation'].S0 = 1200.
    EBM_iceline.S0 = 1200.
    EBM_iceline.integrate_years(5.)
    assert np.all(EBM_iceline.icelat == np.array([-0.,  0.]))

@pytest.mark.fast
def test_float():
    '''Check that we can step forward the model after setting the state
    variable with an ndarray of integers through 2 different methods'''
    from climlab.domain import initial
    from climlab.domain.field import Field
    state = initial.surface_state()
    sfc = climlab.domain.zonal_mean_surface(num_lat=4)
    state.Ts = Field([10,15,15,10],domain=sfc)
    m = climlab.EBM(state=state)
    m.step_forward()
    k = climlab.EBM(num_lat=4)
    k.set_state('Ts', Field([10,15,15,10], domain=k.domains['Ts']))
    k.step_forward()

@pytest.mark.fast
def test_albedo():
    '''Check that we can integrate forward a model after changing the albedo
    subprocess and get the expected icelat'''
    import numpy as np
    m = climlab.EBM()
    m.add_subprocess('albedo', climlab.surface.ConstantAlbedo(state=m.state, **m.param))
    m.subprocess['SW'].albedo = m.subprocess['albedo'].albedo
    m.integrate_years(1)
    m.add_subprocess('albedo', climlab.surface.StepFunctionAlbedo(state=m.state, **m.param))
    m.subprocess['SW'].albedo = m.subprocess['albedo'].albedo
    m.integrate_years(1)
    assert np.all(m.icelat == np.array([-70.,  70.]))
    assert np.all(m.icelat == m.subprocess['albedo'].subprocess['iceline'].icelat)
    #  What is the expected behavior if we swap out a subprocess for another
    #  and they have different diagnostics???
    #m.add_subprocess('albedo', albedo.ConstantAlbedo(state=m.state, **m.param))
    #m.integrate_years(1)
    #assert m.icelat == None

@pytest.mark.fast
def test_analytical():
    '''Check to see if the the numerical solution converges to the analytical
    steady-state solution of the simple EBM with constant albedo'''
    param = {'a0': 0.3,
             'a2': 0.,
             'ai': 0.3,
             's2': -0.48,
             'S0': 1360.,
             'A': 210.,
             'B': 2.,
             'D': 0.55,
             'Tf': -1000., # effectively makes albedo constant
            }
    m = climlab.EBM(**param)
    m.integrate_years(5)
    Tnumerical = np.squeeze(m.Ts)
    delta = param['D']/param['B']
    x = np.sin(np.deg2rad(m.lat))
    Tanalytical = ((1-param['a0'])*param['S0']/4*(1+param['s2']*P2(x)/(1+6*delta))-param['A'])/param['B']
    assert Tnumerical == pytest.approx(Tanalytical, abs=2E-2)

@pytest.mark.fast
def test_moist_EBM_creation():
    '''See if we can swap a moist diffusion module for the dry diffusion
    and just step forward once.'''
    m = climlab.EBM()
    m.remove_subprocess('diffusion')
    diff = climlab.dynamics.MeridionalMoistDiffusion(state=m.state, timestep=m.timestep)
    m.add_subprocess('diffusion', diff)
    m.step_forward()
    assert hasattr(m, 'heat_transport')



from __future__ import division
from __future__ import absolute_import
import numpy as np
from climlab import constants as const
from climlab.radiation.radiation import _Radiation_SW, _Radiation_LW
from .rrtmg_lw import RRTMG_LW
from .rrtmg_sw import RRTMG_SW, nbndsw


class RRTMG(_Radiation_SW, _Radiation_LW):
    '''Container to drive combined LW and SW radiation models.

    For some details about inputs and diagnostics, see the `radiation` module.
    '''
    def __init__(self,
            # GENERAL, used in both SW and LW
            icld = 1,    # Cloud overlap method, 0: Clear only, 1: Random, 2,  Maximum/random] 3: Maximum
            irng = 1,  # more monte carlo stuff
            idrv = 0,  # whether to also calculate the derivative of flux with respect to surface temp
            permuteseed_sw =  150,  # used for monte carlo clouds; must differ from permuteseed_lw by number of subcolumns
            permuteseed_lw =  300,  # learn about these later...
            dyofyr = 0,       # day of the year used to get Earth/Sun distance (if not adjes)
            # CLOUDS, SW see http://www.arm.gov/publications/proceedings/conf16/extended_abs/iacono_mj.pdf
            inflgsw  = 2, # Flag for cloud optical properties
                        # INFLAG = 0 direct specification of optical depths of clouds;
                        #            cloud fraction and cloud optical depth (gray) are
                        #            input for each cloudy layer
                        #        = 1 calculation of combined ice and liquid cloud optical depths (gray)
                        #            as in CCM2; cloud fraction and cloud water path are input for
                        #            each cloudy layer.
                        #        = 2 calculation of separate ice and liquid cloud optical depths, with
                        #            parameterizations determined by values of ICEFLAG and LIQFLAG.
                        #            Cloud fraction, cloud water path, cloud ice fraction, and
                        #            effective ice radius are input for each cloudy layer for all
                        #            parameterizations.  If LIQFLAG = 1, effective liquid droplet radius
                        #            is also needed.
            inflglw  = 2,
            iceflgsw = 1,  # Flag for ice particle specification
                        #             ICEFLAG = 0 the optical depths (gray) due to ice clouds are computed as in CCM3.
                        #                     = 1 the optical depths (non-gray) due to ice clouds are computed as closely as
                        #                         possible to the method in E.E. Ebert and J.A. Curry, JGR, 97, 3831-3836 (1992).
                        #                     = 2 the optical depths (non-gray) due to ice clouds are computed by a method
                        #                         based on the parameterization used in the radiative transfer model Streamer
                        #                         (reference,  J. Key, Streamer User's Guide, Technical Report 96-01] Boston
                        #                         University, 85 pp. (1996)), which is closely related to the parameterization
                        #                         of water clouds due to Hu and Stamnes (see below).
                        #             = 3 the optical depths (non-gray) due to ice clouds are computed by a method
                        # based on the parameterization given in Fu et al., J. Clim.,11,2223-2237 (1998).
                           # specific definition of reic depends on setting of iceflglw:
                           # iceflglw = 0,  ice effective radius, r_ec, (Ebert and Curry, 1992)]
                           #               r_ec must be >= 10.0 microns
                           # iceflglw = 1,  ice effective radius, r_ec, (Ebert and Curry, 1992)]
                           #               r_ec range is limited to 13.0 to 130.0 microns
                           # iceflglw = 2,  ice effective radius, r_k, (Key, Streamer Ref. Manual] 1996)
                           #               r_k range is limited to 5.0 to 131.0 microns
                           # iceflglw = 3,  generalized effective size, dge, (Fu, 1996)]
                           #               dge range is limited to 5.0 to 140.0 microns
                           #               [dge = 1.0315 * r_ec]

            iceflglw = 1,
            liqflgsw = 1,  # Flag for liquid droplet specification
                        # LIQFLAG = 0 the optical depths (gray) due to water clouds are computed as in CCM3.
                        #         = 1 the optical depths (non-gray) due to water clouds are computed by a method
                        #             based on the parameterization of water clouds due to Y.X. Hu and K. Stamnes,
                        #             J. Clim., 6, 728-742 (1993).
            liqflglw = 1,
            tauc_sw = 0.,  # In-cloud optical depth
            tauc_lw = 0.,  # in-cloud optical depth
            ssac_sw = 0.,  # In-cloud single scattering albedo
            asmc_sw = 0.,  # In-cloud asymmetry parameter
            fsfc_sw = 0.,  # In-cloud forward scattering fraction (delta function pointing forward "forward peaked scattering")
            # AEROSOLS
            tauaer_sw = 0.,   # Aerosol optical depth (iaer=10 only), Dimensions,  (ncol,nlay,nbndsw)] #  (non-delta scaled)
            ssaaer_sw = 0.,   # Aerosol single scattering albedo (iaer=10 only), Dimensions,  (ncol,nlay,nbndsw)] #  (non-delta scaled)
            asmaer_sw = 0.,   # Aerosol asymmetry parameter (iaer=10 only), Dimensions,  (ncol,nlay,nbndsw)] #  (non-delta scaled)
            ecaer_sw  = 0.,   # Aerosol optical depth at 0.55 micron (iaer=6 only), Dimensions,  (ncol,nlay,naerec)] #  (non-delta scaled)
            tauaer_lw = 0.,   # Aerosol optical depth at mid-point of LW spectral bands
            # new arguments for RRTMG_SW version 4.0
            isolvar = 0,    # ! Flag for solar variability method
                            # !   -1 = (when scon .eq. 0.0): No solar variability
                            # !        and no solar cycle (Kurucz solar irradiance
                            # !        of 1368.22 Wm-2 only);
                            # !        (when scon .ne. 0.0): Kurucz solar irradiance
                            # !        scaled to scon and solar variability defined
                            # !        (optional) by setting non-zero scale factors
                            # !        for each band in bndsolvar
                            # !    0 = (when SCON .eq. 0.0): No solar variability
                            # !        and no solar cycle (NRLSSI2 solar constant of
                            # !        1360.85 Wm-2 for the 100-50000 cm-1 spectral
                            # !        range only), with facular and sunspot effects
                            # !        fixed to the mean of Solar Cycles 13-24;
                            # !        (when SCON .ne. 0.0): No solar variability
                            # !        and no solar cycle (NRLSSI2 solar constant of
                            # !        1360.85 Wm-2 for the 100-50000 cm-1 spectral
                            # !        range only), is scaled to SCON
                            # !    1 = Solar variability (using NRLSSI2  solar
                            # !        model) with solar cycle contribution
                            # !        determined by fraction of solar cycle
                            # !        with facular and sunspot variations
                            # !        fixed to their mean variations over the
                            # !        average of Solar Cycles 13-24;
                            # !        two amplitude scale factors allow
                            # !        facular and sunspot adjustments from
                            # !        mean solar cycle as defined by indsolvar
                            # !    2 = Solar variability (using NRLSSI2 solar
                            # !        model) over solar cycle determined by
                            # !        direct specification of Mg (facular)
                            # !        and SB (sunspot) indices provided
                            # !        in indsolvar (scon = 0.0 only)
                            # !    3 = (when scon .eq. 0.0): No solar variability
                            # !        and no solar cycle (NRLSSI2 solar irradiance
                            # !        of 1360.85 Wm-2 only);
                            # !        (when scon .ne. 0.0): NRLSSI2 solar irradiance
                            # !        scaled to scon and solar variability defined
                            # !        (optional) by setting non-zero scale factors
                            # !        for each band in bndsolvar
            indsolvar = np.zeros(2),      # Facular and sunspot amplitude scale factors (isolvar=1),
                                          # or Mg and SB indices (isolvar=2)
            bndsolvar = np.zeros(nbndsw), # Solar variability scale factors for each shortwave band
            solcycfrac = 0.,              # Fraction of averaged solar cycle (0-1) at current time (isolvar=1)

            **kwargs):
        super(RRTMG, self).__init__(**kwargs)

        # Remove specific inputs from kwargs dictionary.
        #  We want any changes implemented in the parent __init__ method to be preserved here
        remove_list = ['absorber_vmr','cldfrac','clwp','ciwp','r_liq','r_ice',
                       'emissivity','aldif','aldir','asdif','asdir','S0','coszen',
                       'eccentricity_factor','insolation',]
        for item in remove_list:
            if item in kwargs:
                ignored = kwargs.pop(item)

        LW = RRTMG_LW(absorber_vmr = self.absorber_vmr,
                     cldfrac = self.cldfrac,
                     clwp = self.clwp,
                     ciwp = self.ciwp,
                     r_liq = self.r_liq,
                     r_ice = self.r_ice,
                     icld = icld,
                     irng = irng,
                     idrv = idrv,
                     permuteseed = permuteseed_lw,
                     emissivity = self.emissivity,
                     inflglw = inflglw,
                     iceflglw = iceflglw,
                     liqflglw = liqflglw,
                     tauc = tauc_lw,
                     tauaer = tauaer_lw,
                     **kwargs)
        SW = RRTMG_SW(absorber_vmr = self.absorber_vmr,
                     cldfrac = self.cldfrac,
                     clwp = self.clwp,
                     ciwp = self.ciwp,
                     r_liq = self.r_liq,
                     r_ice = self.r_ice,
                     icld = icld,
                     irng = irng,
                     permute = permuteseed_sw,
                     aldif = self.aldif,
                     aldir = self.aldir,
                     asdif = self.asdif,
                     asdir = self.asdir,
                     S0 = self.S0,
                     coszen = self.coszen,
                     eccentricity_factor = self.eccentricity_factor,
                     dyofyr = dyofyr,
                     insolation = self.insolation,
                     inflgsw = inflgsw,
                     iceflgsw = iceflgsw,
                     tauc = tauc_sw,
                     ssac = ssac_sw,
                     asmc = asmc_sw,
                     fsfc = fsfc_sw,
                     tauaer = tauaer_sw,
                     ssaaer = ssaaer_sw,
                     asmaer = asmaer_sw,
                     ecaer = ecaer_sw,
                     isolvar = isolvar,
                     indsolvar = indsolvar,
                     bndsolvar = bndsolvar,
                     solcycfrac = solcycfrac,
                     **kwargs)
        self.add_subprocess('SW', SW)
        self.add_subprocess('LW', LW )

        self.add_input('icld', icld)
        self.add_input('irng', irng)
        self.add_input('idrv', idrv)
        self.add_input('permuteseed_sw', permuteseed_sw)
        self.add_input('permuteseed_lw', permuteseed_lw)
        self.add_input('dyofyr', dyofyr)
        self.add_input('inflgsw', inflgsw)
        self.add_input('inflglw', inflglw)
        self.add_input('iceflgsw', iceflgsw)
        self.add_input('iceflglw', iceflglw)
        self.add_input('liqflgsw', liqflgsw)
        self.add_input('liqflglw', liqflglw)
        self.add_input('tauc_sw', tauc_sw)
        self.add_input('tauc_lw', tauc_lw)
        self.add_input('ssac_sw', ssac_sw)
        self.add_input('asmc_sw', asmc_sw)
        self.add_input('fsfc_sw', fsfc_sw)
        self.add_input('tauaer_sw', tauaer_sw)
        self.add_input('ssaaer_sw', ssaaer_sw)
        self.add_input('asmaer_sw', asmaer_sw)
        self.add_input('ecaer_sw', ecaer_sw)
        self.add_input('tauaer_lw', tauaer_lw)
        self.add_input('isolvar', isolvar)
        self.add_input('indsolvar', indsolvar)
        self.add_input('bndsolvar', bndsolvar)
        self.add_input('solcycfrac', solcycfrac)



from __future__ import division
import numpy as np
import climlab
import pytest
from climlab.tests.xarray_test import to_xarray

# The fixtures are reusable pieces of code to set up the input to the tests.
# Without fixtures, we would have to do a lot of cutting and pasting
# I inferred which fixtures to use from the notebook
# Latitude-dependent grey radiation.ipynb
@pytest.fixture()
def model():
    return climlab.BandRCModel()

@pytest.fixture()
def diffmodel():
    ''' 2D radiative-convective model with band radiation including water vapor,
    fixed relative humidity, meridional heat transport (diffusion) and convective adjustment.
    '''
    diffmodel = climlab.BandRCModel(num_lev=30, num_lat=90)
    insolation = climlab.radiation.AnnualMeanInsolation(domains=diffmodel.Ts.domain)
    diffmodel.add_subprocess('insolation', insolation)
    diffmodel.subprocess.SW.flux_from_space = insolation.insolation
    # thermal diffusivity in W/m**2/degC
    D = 0.05
    # meridional diffusivity in 1/s
    K = D / diffmodel.Tatm.domain.heat_capacity[0]
    d = climlab.dynamics.MeridionalDiffusion(K=K,
                state={'Tatm': diffmodel.state['Tatm']},
                **diffmodel.param)
    diffmodel.add_subprocess('diffusion', d)
    return diffmodel

@pytest.fixture()
def diffmodel_surfflux(diffmodel):
    '''Explicit surface sensible and latent heat fluxes.'''
    diffmodel_surfflux = climlab.process_like(diffmodel)
    # process models for surface heat fluxes
    shf = climlab.surface.SensibleHeatFlux(state=diffmodel_surfflux.state, Cd=0.5E-3)
    lhf = climlab.surface.LatentHeatFlux(state=diffmodel_surfflux.state, Cd=0.5E-3)
    # set the water vapor input field for LHF
    lhf.q = diffmodel_surfflux.q
    diffmodel_surfflux.add_subprocess('SHF', shf)
    diffmodel_surfflux.add_subprocess('LHF', lhf)
    #  Convective adjustment for atmosphere only
    diffmodel_surfflux.remove_subprocess('convective adjustment')
    conv = climlab.convection.ConvectiveAdjustment(state={'Tatm':diffmodel_surfflux.state['Tatm']},
                                **diffmodel_surfflux.param)
    diffmodel_surfflux.add_subprocess('convective adjustment', conv)
    return diffmodel_surfflux


# helper for a common test pattern
def _check_minmax(array, amin, amax):
    return (np.allclose(array.min(), amin) and
            np.allclose(array.max(), amax))

@pytest.mark.fast
def test_model_creation(model):
    """Just make sure we can create a model."""
    assert len(model.Tatm)==30

@pytest.mark.fast
def test_diffmodel(diffmodel):
    """Check that we can integrate the model with diffusion."""
    diffmodel.step_forward()
    diffmodel.integrate_days(2)
    #  Would be better to have these tests evaluate a numerical condition
    #  Test the xarray interface
    to_xarray(diffmodel)

@pytest.mark.fast
def test_diffmodel_surfflux(diffmodel_surfflux):
    """Check that we can integrate the model with diffusion."""
    diffmodel_surfflux.step_forward()
    diffmodel_surfflux.integrate_days(2)

@pytest.mark.slow
def test_integrate_years(model):
    """Check that we can integrate forward the model and get the expected
    surface temperature and water vapor.
    Also check the climate sensitivity to doubling CO2."""
    model.step_forward()
    model.integrate_years(2)
    Ts = model.Ts.copy()
    assert np.isclose(Ts, 275.43383753)
    assert _check_minmax(model.q, 5.E-6, 3.23764447e-03)
    model.absorber_vmr['CO2'] *= 2.
    model.integrate_years(2)
    assert np.isclose(model.Ts - Ts, 3.180993)



from __future__ import division
from climlab import constants as const
from climlab.process.energy_budget import EnergyBudget


class Boltzmann(EnergyBudget):
    r"""A class for black body radiation.

    Implements a radiation subprocess which computes longwave radiation
    with the Stefan-Boltzmann law for black/grey body radiation.

    According to the Stefan Boltzmann law the total power radiated from an
    object with surface area :math:`A` and temperature :math:`T` (in unit Kelvin)
    can be written as

    .. math::

        P = A \varepsilon \sigma T^4

    where :math:`\varepsilon` is the emissivity of the body.

    As the :class:`~climlab.process.energy_budget.EnergyBudget` of the
    Energy Balance Model is accounted in unit :math:`\textrm{energy} / \textrm{area}`
    (:math:`\textrm{W}/ \textrm{m}^2`)
    the energy budget equation looks like this:

    .. math::

        C \frac{dT}{dt} = R\downarrow - R\uparrow - H  \n

    The :class:`Boltzmann` radiation subprocess represents the outgoing radiation
    :math:`R\uparrow` which then can be written as

    .. math::

        R\uparrow = \varepsilon \sigma T^4

    with state variable :math:`T`.

    **Initialization parameters** \n

    An instance of ``Boltzmann`` is initialized with the following
    arguments:

    :param float eps:   emissivity of the planet's surface which is the
                        effectiveness in emitting energy as thermal radiation
                        [default: 0.65]
    :param float tau:   transmissivity of the planet's atmosphere which is the
                        effectiveness in transmitting the longwave radiation
                        emitted from the surface [default: 0.95]

    **Object attributes** \n

    During initialization both arguments described above are created as object
    attributes which calls their setter function (see below).

    :ivar float eps:            calls the setter function of :func:`eps`
    :ivar float tau:            calls the setter function of :func:`tau`
    :ivar dict diagnostics:     the subprocess's diagnostic dictionary
                                ``self.diagnostic`` is initialized
                                through calling
                                ``self.add_diagnostic('OLR', 0. * self.Ts)``
    :ivar Field OLR:            the subprocess attribute ``self.OLR`` is
                                created with correct dimensions

    :Example:

        Replacing an the regular AplusBT subprocess in an energy balance model::

            >>> import climlab
            >>> from climlab.radiation.Boltzmann import Boltzmann

            >>> # creating EBM model
            >>> model = climlab.EBM()

            >>> print model

        .. code-block:: none
            :emphasize-lines: 7

            climlab Process of type <class 'climlab.model.ebm.EBM'>.
            State variables and domain shapes:
              Ts: (90, 1)
            The subprocess tree:
            top: <class 'climlab.model.ebm.EBM'>
               diffusion: <class 'climlab.dynamics.diffusion.MeridionalDiffusion'>
               LW: <class 'climlab.radiation.AplusBT.AplusBT'>
               albedo: <class 'climlab.surface.albedo.StepFunctionAlbedo'>
                  iceline: <class 'climlab.surface.albedo.Iceline'>
                  cold_albedo: <class 'climlab.surface.albedo.ConstantAlbedo'>
                  warm_albedo: <class 'climlab.surface.albedo.P2Albedo'>
               insolation: <class 'climlab.radiation.insolation.P2Insolation'>

        ::

            >>> #  creating and adding albedo feedback subprocess
            >>> LW_boltz = Boltzmann(eps=0.69, tau=0.98, state=model.state, **model.param)

            >>> # overwriting old 'LW' subprocess with same name
            >>> model.add_subprocess('LW', LW_boltz)

            >>> print model

        .. code-block:: none
            :emphasize-lines: 7

            climlab Process of type <class 'climlab.model.ebm.EBM'>.
            State variables and domain shapes:
              Ts: (90, 1)
            The subprocess tree:
            top: <class 'climlab.model.ebm.EBM'>
               diffusion: <class 'climlab.dynamics.diffusion.MeridionalDiffusion'>
               LW: <class 'climlab.radiation.Boltzmann.Boltzmann'>
               albedo: <class 'climlab.surface.albedo.StepFunctionAlbedo'>
                  iceline: <class 'climlab.surface.albedo.Iceline'>
                  cold_albedo: <class 'climlab.surface.albedo.ConstantAlbedo'>
                  warm_albedo: <class 'climlab.surface.albedo.P2Albedo'>
               insolation: <class 'climlab.radiation.insolation.P2Insolation'>

    """
    # implemented by m-kreuzer
    def __init__(self, eps= 0.65, tau=0.95, **kwargs):
        super(Boltzmann, self).__init__(**kwargs)
        self.eps = eps
        self.tau = tau
        self.add_diagnostic('OLR', 0. * self.Ts)

    @property
    def eps(self):
        """Property of emissivity parameter.

        :getter:    Returns the albedo value which is stored in attribute
                    ``self._eps``
        :setter:    * sets the emissivity which is addressed as ``self._eps``
                      to the new value
                    * updates the parameter dictionary ``self.param['eps']``
        :type:      float

        """
        return self._eps
    @eps.setter
    def eps(self, value):
        self._eps = value
        self.param['eps'] = value

    @property
    def tau(self):
        """Property of the transmissivity parameter.

        :getter:    Returns the albedo value which is stored in attribute
                    ``self._tau``
        :setter:    * sets the emissivity which is addressed as ``self._tau``
                      to the new value
                    * updates the parameter dictionary ``self.param['tau']``
        :type:      float

        """
        return self._tau
    @tau.setter
    def tau(self, value):
        self._tau = value
        self.param['tau'] = value

#    def emission(self):
#        """Calculates the Outgoing Longwave Radiation (OLR) of the Boltzmann
#        radiation subprocess.
#
#        **Object attributes** \n
#
#        During method execution following object attribute is modified:
#
#        :ivar float OLR:            the described formula is calculated and the
#                                    result stored in the project attribute ``self.OLR``
#        :ivar dict diagnostics:     the same result is written in ``diagnostics``
#                                    dictionary with the key ``'OLR'``
#
#        .. warning::
#
#            This currently works only for a single state variable!
#
#        """
#        for varname, value in self.state.iteritems():
#            flux = self.eps * self.tau * const.sigma * (value + const.tempCtoK)**4.
#            self.OLR = flux
#            self.diagnostics['OLR'] = self.OLR

    def _compute_emission(self):
        for varname, value in self.state.items():
            flux = self.eps * self.tau * const.sigma * (value + const.tempCtoK)**4.
            self.OLR[:] = flux


    def _compute_heating_rates(self):
        """Computes energy flux convergences to get heating rates in :math:`W/m^2`.

        """
        self._compute_emission()
        for varname, value in self.state.items():
            self.heating_rate[varname] = -self.OLR



from __future__ import division
import numpy as np
import climlab
from climlab.utils import thermo
import pytest


@pytest.mark.fast
def test_thermo():
    '''Basic single value tests for the thermodynamic routines.'''
    assert np.isclose(thermo.potential_temperature(250., 500.), 304.783)
    assert np.isclose(thermo.theta(250., 500.), 304.783)

    assert np.isclose(thermo.temperature_from_potential(300., 500.), 246.076)
    assert np.isclose(thermo.T(300., 500.), 246.076)

    assert np.isclose(thermo.clausius_clapeyron(300.), 35.345)

    assert np.isclose(thermo.qsat(300., 1000.), 0.02227839)

    assert np.isclose(thermo.estimated_inversion_strength(300., 290.), 5.3605345)
    assert np.isclose(thermo.EIS(300., 290.), 5.3605345)

    assert np.isclose(thermo.blackbody_emission(300.), 459.3)

@pytest.mark.fast
def test_thermo_domain():
    '''Can we call qsat, etc on a multi-dim climlab state temperature object?'''
    state = climlab.column_state(num_lev = 30, num_lat=3)
    T = state.Tatm
    p = T.domain.lev.points
    thermo.clausius_clapeyron(T)
    thermo.qsat(T, p)
    thermo.pseudoadiabat(T, p)
    thermo.blackbody_emission(T)
